"""Contains PyDSS utility functions"""

import json

from PyDSS.pydss_project import PyDssProject


def detect_convergence_problems(project_path):
    """Detects convergence problems in a PyDSS run.

    Parameters
    ----------
    project_path : str

    Returns
    -------
    list
        list of dicts with keys: 'scenario', 'priority', 'step'

    """
    problems = []
    project = PyDssProject.load_project(project_path)
    project_name = project.simulation_config.project.active_project
    for scenario in project.list_scenario_names():
        log_file = f"Logs/{project_name}__{scenario}__reports.log"
        problems += _detect_convergence_problems(project.fs_interface.read_file(log_file))

    return problems


def _detect_convergence_problems(text):
    problems = []
    for line in text.splitlines():
        data = json.loads(line)
        if data["Report"] == "Convergence":
            problems.append(data)

    return problems



"""Defines Option objects."""

class Option:
    """Option class used to store analysis option values"""

    def __init__(self, name, default, value=None):
        self.name = name
        self.default = default
        self.value = value



import os
import sys

from disco.enums import SimulationType
from disco.pipelines.base import PipelineTemplate
from disco.pipelines.enums import TemplateSection
from disco.sources.base import FORMAT_FILENAME


def get_default_pipeline_template(simulation_type):
    """Return the default pipeline template file"""
    if isinstance(simulation_type, SimulationType):
        simulation_type = simulation_type.value
    filename = f"{simulation_type}-default-template.toml"
    template_file = os.path.join(os.path.dirname(__file__), "template", filename)
    template = PipelineTemplate(template_file)
    return template


def check_hpc_config(template_file):
    template = PipelineTemplate(template_file)
    for section in TemplateSection:
        try:
            submitter_params = template.get_submitter_params(section)
        except KeyError:
            continue
        
        hpc_config =  submitter_params.get("hpc_config")
        if isinstance(hpc_config, dict) and hpc_config["hpc_type"] == "local":
            return

        assert isinstance(hpc_config, str)
        hpc_config_file = hpc_config
        if not hpc_config_file or not os.path.exists(hpc_config_file):
            print(f"{hpc_config_file} does not exist, please run 'jade config hpc' to create.")
            sys.exit(1)


def ensure_jade_pipeline_output_dir(text):
    """Replace JADE_PIPELINE_OUTPUT_DIR by using its value."""
    if "JADE_PIPELINE_OUTPUT_DIR" not in text:
        return text
    if "JADE_PIPELINE_OUTPUT_DIR" not in os.environ:
        raise KeyError(f"Environment variable JADE_PIPELINE_OUTPUT_DIR not found.")
    return text.replace("$JADE_PIPELINE_OUTPUT_DIR", os.environ["JADE_PIPELINE_OUTPUT_DIR"])



from sqlalchemy import Column, ForeignKey
from sqlalchemy import types
from sqlalchemy.engine import create_engine as _create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class Task(Base):
    __tablename__ = "task"
    id = Column(types.String(length=36), primary_key=True)
    name = Column(types.String(length=256))
    inputs = Column(types.Text())
    output = Column(types.Text())
    image_version = Column(types.String(length=30), nullable=True, default=None)
    disco_version = Column(types.String(length=30), nullable=True, default=None)
    jade_version = Column(types.String(length=30), nullable=True, default=None)
    pydss_version = Column(types.String(length=30), nullable=True, default=None)
    opendssdirect_version = Column(types.String(length=30), nullable=True, default=None)
    opendss_version = Column(types.Text(), nullable=True, default=None)
    notes = Column(types.Text, nullable=True, default=None)
    creation_time = Column(types.DateTime)


class Job(Base):
    __tablename__ = "job"
    id = Column(types.String(length=36), primary_key=True)
    task_id = Column(types.String(length=36), ForeignKey("task.id"))
    name = Column(types.String(length=256))
    project_path = Column(types.Text(), nullable=True)
    return_code = Column(types.Integer, nullable=True)
    status = Column(types.String(length=20), nullable=True)
    exec_time_s = Column(types.Float, nullable=True)
    completion_time = Column(types.DateTime)
    task = relationship("Task", primaryjoin="Job.task_id == Task.id")


class Scenario(Base):
    __tablename__ = "scenario"
    id = Column(types.String(length=36), primary_key=True)
    job_id = Column(types.String(length=36), ForeignKey("job.id"))
    simulation_type = Column(types.String(length=30))
    name = Column(types.String(length=30))
    start_time = Column(types.DateTime, nullable=True)
    end_time = Column(types.DateTime, nullable=True)
    job = relationship("Job", primaryjoin="Scenario.job_id == Job.id")


class SnapshotTimePoints(Base):
    __tablename__ = "snapshot_time_points"
    id = Column(types.String(length=36), primary_key=True)
    job_id = Column(types.String(length=36), ForeignKey("job.id"))
    max_pv_load_ratio = Column(types.DateTime)
    max_load = Column(types.DateTime)
    daytime_min_load = Column(types.DateTime)
    pv_minus_load = Column(types.DateTime)
    job = relationship("Job", primaryjoin="SnapshotTimePoints.job_id == Job.id")


class Report(Base):
    __tablename__ = "report"
    id = Column(types.String(length=36), primary_key=True)
    task_id = Column(types.String(length=36), ForeignKey("task.id"))
    file_name = Column(types.String(length=256))
    file_path = Column(types.Text())
    file_size = Column(types.Integer)
    creation_time = Column(types.DateTime)
    task = relationship("Task", primaryjoin="Report.task_id == Task.id")


class FeederHead(Base):
    __tablename__ = "feeder_head"
    id = Column(types.String(length=36), primary_key=True)
    report_id = Column(types.String(length=36), ForeignKey("report.id"))
    job_id = Column(types.String(length=36), ForeignKey("job.id"), nullable=True)
    name = Column(types.String(length=256))
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=30), nullable=True)
    sample = Column(types.Float, nullable=True)
    penetration_level = Column(types.Float, nullable=True)
    scenario = Column(types.String(length=128))
    line = Column(types.String(length=128))
    loading = Column(types.Float)
    load_kw = Column(types.Float)
    load_kvar = Column(types.Float)
    reverse_power_flow = Column(types.Boolean)
    job = relationship("Job", primaryjoin="FeederHead.job_id == Job.id")
    report = relationship("Report", primaryjoin="FeederHead.report_id == Report.id")


class FeederLosses(Base):
    __tablename__ = "feeder_losses"
    id = Column(types.String(length=36), primary_key=True)
    report_id = Column(types.String(length=36), ForeignKey("report.id"))
    job_id = Column(types.String(length=36), ForeignKey("job.id"), nullable=True)
    name = Column(types.String(length=256))
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=30), nullable=True)
    sample = Column(types.Float, nullable=True)
    penetration_level = Column(types.Float, nullable=True)
    scenario = Column(types.String(length=128))
    total_losses_kwh = Column(types.Float)
    line_losses_kwh = Column(types.Float)
    transformer_losses_kwh = Column(types.Float)
    total_load_demand_kwh = Column(types.Float)
    job = relationship("Job", primaryjoin="FeederLosses.job_id == Job.id")
    report = relationship("Report", primaryjoin="FeederLosses.report_id == Report.id")


class Metadata(Base):
    __tablename__ = "metadata"
    id = Column(types.String(length=36), primary_key=True)
    report_id = Column(types.String(length=36), ForeignKey("report.id"))
    job_id = Column(types.String(length=36), ForeignKey("job.id"), nullable=True)
    name = Column(types.String(length=256))
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=30), nullable=True)
    sample = Column(types.Float, nullable=True)
    penetration_level = Column(types.Float, nullable=True)
    scenario = Column(types.String(length=128))
    pct_pv_to_load_ratio = Column(types.Float)
    pv_capacity_kw = Column(types.Float)
    load_capacity_kw = Column(types.Float)
    job = relationship("Job", primaryjoin="Metadata.job_id == Job.id")
    report = relationship("Report", primaryjoin="Metadata.report_id == Report.id")


class VoltageMetrics(Base):
    __tablename__ = "voltage_metrics"
    id = Column(types.String(length=36), primary_key=True)
    report_id = Column(types.String(length=36), ForeignKey("report.id"))
    job_id = Column(types.String(length=36), ForeignKey("job.id"), nullable=True)
    name = Column(types.String(length=256))
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=30), nullable=True)
    sample = Column(types.Float, nullable=True)
    penetration_level = Column(types.Float, nullable=True)
    scenario = Column(types.String(length=128))
    node_type = Column(types.String(length=10))
    num_nodes_any_outside_ansi_b = Column(types.Integer)
    num_time_points_with_ansi_b_violations = Column(types.Integer)
    voltage_duration_between_ansi_a_and_b_minutes = Column(types.Integer)
    max_per_node_voltage_duration_outside_ansi_a_minutes = Column(types.Integer)
    moving_average_voltage_duration_outside_ansi_a_minutes = Column(types.Integer)
    num_nodes_always_inside_ansi_a = Column(types.Integer)
    num_nodes_any_outside_ansi_a_always_inside_ansi_b = Column(types.Integer)
    min_voltage = Column(types.Float)
    max_voltage = Column(types.Float)
    job = relationship("Job", primaryjoin="VoltageMetrics.job_id == Job.id")
    report = relationship("Report", primaryjoin="VoltageMetrics.report_id == Report.id")


class ThermalMetrics(Base):
    __tablename__ = "thermal_metrics"
    id = Column(types.String(length=36), primary_key=True)
    report_id = Column(types.String(length=36), ForeignKey("report.id"))
    job_id = Column(types.String(length=36), ForeignKey("job.id"), nullable=True)
    name = Column(types.String(length=256))
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=30), nullable=True)
    sample = Column(types.Float, nullable=True)
    penetration_level = Column(types.Float, nullable=True)
    scenario = Column(types.String(length=128))
    line_max_instantaneous_loading_pct = Column(types.Float)
    line_max_moving_average_loading_pct = Column(types.Float)
    line_window_size_hours = Column(types.Integer)
    line_num_time_points_with_instantaneous_violations = Column(types.Integer)
    line_num_time_points_with_moving_average_violations = Column(types.Integer)
    line_instantaneous_threshold = Column(types.Float)
    line_moving_average_threshold = Column(types.Float)
    transformer_max_instantaneous_loading_pct = Column(types.Float)
    transformer_max_moving_average_loading_pct = Column(types.Float)
    transformer_window_size_hours = Column(types.Integer)
    transformer_num_time_points_with_instantaneous_violations = Column(types.Integer)
    transformer_num_time_points_with_moving_average_violations = Column(types.Integer)
    transformer_instantaneous_threshold = Column(types.Integer)
    transformer_moving_average_threshold = Column(types.Integer)
    job = relationship("Job", primaryjoin="ThermalMetrics.job_id == Job.id")
    report = relationship("Report", primaryjoin="ThermalMetrics.report_id == Report.id")


class HostingCapacity(Base):
    __tablename__ = "hosting_capacity"
    id = Column(types.String(length=36), primary_key=True)
    task_id = Column(types.String(length=36), ForeignKey("task.id"))
    hc_type = Column(types.String(length=20))
    metric_type = Column(types.String(length=20))
    feeder = Column(types.String(length=256))
    scenario = Column(types.String(length=256))
    time_point = Column(types.String(length=60), nullable=True, default=None)
    min_hc_pct = Column(types.Float)
    max_hc_pct = Column(types.Float)
    min_hc_kw = Column(types.Float)
    max_hc_kw = Column(types.Float)
    task = relationship("Task", primaryjoin="HostingCapacity.task_id == Task.id")


class PvDistances(Base):
    __tablename__ = "pv_distances"
    job_name = Column(types.String(length=256), primary_key=True)
    substation = Column(types.String(length=256))
    feeder = Column(types.String(length=256))
    placement = Column(types.String(length=256), nullable=True)
    sample = Column(types.Integer, nullable=True)
    penetration_level = Column(types.Integer, nullable=True)
    weighted_average_pv_distance = Column(types.Float)
    option = Column(types.String(length=128))


def create_engine(database):
    engine = _create_engine("sqlite:///" + database)
    return engine


def create_database(database):
    engine = create_engine(database)
    Base.metadata.create_all(engine)



import logging
import os
from pathlib import Path

import click


from jade.extensions.generic_command.generic_command_configuration import (
    GenericCommandConfiguration,
)
from jade.extensions.generic_command.generic_command_parameters import (
    GenericCommandParameters,
)
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.utils.utils import load_data

from disco.pydss.prescreen_pv_penetration_levels import (
    create_job_key,
    run_prescreen,
    JobKey,
    PRESCREEN_JOBS_OUTPUT,
)


logger = logging.getLogger(__name__)


@click.group()
@click.argument("config_file")
@click.pass_context
def prescreen_pv_penetration_levels(ctx, config_file):
    """Prescreen jobs to find the highest passing PV penetration level for each sample."""
    ctx.obj = create_config_from_file(config_file)


@click.command()
@click.option(
    "-c",
    "--config-file",
    default="prescreen_config.json",
    show_default=True,
    help="Destination config file",
)
@click.pass_context
@click.pass_obj
def create(config, ctx, config_file):
    """Create prescreen jobs from a JADE config file."""
    jobs = set()
    for job in config.iter_pydss_simulation_jobs():
        if not job.model.is_base_case:
            jobs.add(create_job_key(job))

    src_config_file = ctx.parent.params["config_file"]
    dst_config = GenericCommandConfiguration()
    for key in jobs:
        cmd = (
            f"disco prescreen-pv-penetration-levels {src_config_file} run "
            f"--substation={key.substation} --feeder={key.feeder} --placement={key.placement} "
            f"--sample={key.sample}"
        )
        job = GenericCommandParameters(command=cmd, append_output_dir=True)
        dst_config.add_job(job)

    dst_config.dump(config_file)
    print(f"Wrote jobs to {config_file}")


@click.command()
@click.option(
    "-s",
    "--substation",
    required=True,
)
@click.option(
    "-f",
    "--feeder",
    default="None",
    required=False,
)
@click.option(
    "-p",
    "--placement",
    required=True,
)
@click.option(
    "-d",
    "--sample",
    required=True,
    type=int,
)
@click.option(
    "-o",
    "--jade-runtime-output",
    default="output",
    help="output directory",
)
@click.option(
    "--verbose",
    default=False,
    is_flag=True,
    show_default=True,
    help="Enable debug logging",
)
@click.pass_context
@click.pass_obj
def run(config, ctx, substation, feeder, placement, sample, jade_runtime_output, verbose):
    """Run a bisect on the penetration levels for a sample."""
    prescreen_jobs_output = Path(jade_runtime_output) / PRESCREEN_JOBS_OUTPUT
    os.makedirs(prescreen_jobs_output, exist_ok=True)
    filename = os.path.join(
        prescreen_jobs_output,
        "__".join((substation, feeder, placement, str(sample))) + ".log",
    )
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging("disco", filename, console_level=logging.WARNING, packages=["disco"])

    src_config_file = ctx.parent.params["config_file"]
    run_prescreen(
        config,
        src_config_file,
        substation,
        feeder,
        placement,
        sample,
        prescreen_jobs_output,
    )


@click.command()
@click.argument("output_dir")
@click.option(
    "-c",
    "--config-file",
    default="filtered_config.json",
    help="config file with only prescreened jobs",
)
@click.pass_context
@click.pass_obj
def filter_config(config, ctx, config_file, output_dir):
    """Filter the source config file with the prescreening results."""
    os.makedirs(output_dir, exist_ok=True)
    filename = Path(output_dir) / "filter_prescreened_jobs.log"
    setup_logging("disco", filename, console_level=logging.WARNING, packages=["disco"])
    src_config_file = ctx.parent.params["config_file"]
    prescreen_jobs_output = Path(output_dir) / PRESCREEN_JOBS_OUTPUT
    highest_passing_levels = {}
    for filename in prescreen_jobs_output.glob("*.toml"):
        data = load_data(filename)
        key = JobKey(
            substation=data["substation"],
            feeder=data["feeder"],
            placement=data["placement"],
            sample=data["sample"],
        )
        highest_passing_levels[key] = data["highest_passing_penetration_level"]

    jobs_to_remove = []
    for job in config.iter_pydss_simulation_jobs(exclude_base_case=True):
        key = create_job_key(job)
        highest_passing_level = highest_passing_levels.get(key)
        if highest_passing_level is None:
            logger.warning("Skipping job %s because there is no passing penetration level", key)
            jobs_to_remove.append(job)
        elif job.model.deployment.project_data["penetration_level"] > highest_passing_level:
            jobs_to_remove.append(job)

    for job in jobs_to_remove:
        config.remove_job(job)

    config.dump(config_file)
    num = len(jobs_to_remove)
    logger.info(
        "Created %s by filtering %s job(s) that were expected to fail from %s.",
        config_file,
        num,
        src_config_file,
    )


prescreen_pv_penetration_levels.add_command(create)
prescreen_pv_penetration_levels.add_command(run)
prescreen_pv_penetration_levels.add_command(filter_config)



"""Main CLI command for disco."""

import logging

import click

from disco.cli.compute_snapshot_hosting_capacity import compute_snapshot_hosting_capacity
from disco.cli.compute_snapshot_impact_analysis import compute_snapshot_impact_analysis
from disco.cli.compute_time_series_hosting_capacity import compute_time_series_hosting_capacity
from disco.cli.compute_time_series_impact_analysis import compute_time_series_impact_analysis
from disco.cli.make_cba_tables import make_cba_tables
from disco.cli.make_upgrade_tables import make_upgrade_tables
from disco.cli.cba_post_process import cba_post_process
from disco.cli.compute_cba import compute_cba


logger = logging.getLogger(__name__)


@click.group()
def cli():
    """DISCO internal commands"""


cli.add_command(compute_snapshot_hosting_capacity)
cli.add_command(compute_snapshot_impact_analysis)
cli.add_command(compute_time_series_hosting_capacity)
cli.add_command(compute_time_series_impact_analysis)
cli.add_command(make_cba_tables)
cli.add_command(make_upgrade_tables)
cli.add_command(cba_post_process)
cli.add_command(compute_cba)



import abc
import json
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union, Set

from pydantic.v1 import validator, root_validator, Field, BaseModel
from pydantic.v1.types import DirectoryPath, FilePath

from jade.utils.utils import ExtendedJSONEncoder, standardize_timestamp
from PyDSS.common import ControllerType
from PyDSS.registry import Registry

from disco.enums import SimulationType
from disco.models.utils import SchemaDict
from disco.pydss.pydss_configuration_base import DEFAULT_CONTROLLER_CONFIGS


DISCO_CONTROLLER_NAMES = [
    controller["name"]
    for controller in DEFAULT_CONTROLLER_CONFIGS
]

registered_pydss_controllers = defaultdict(set)


class DiscoBaseModel(BaseModel):
    """Base input model for DISCO types."""

    @classmethod
    def schema_json(
        cls,
        #model_type: str = "SnapshotImpactAnalysisModel",
        by_alias: bool = True,
        indent: int = 2,
    ) -> str:
        data = cls.schema(by_alias=by_alias)
        return json.dumps(data, indent=indent, cls=ExtendedJSONEncoder)

    @classmethod
    def example(
        cls,
        in_list: bool = True,
    ) -> dict:
        """Create an data example for an inputs model"""
        data = {}
        schema_dict = SchemaDict(super().schema())
        for field in cls.__fields__.values():
            if field.name == "model_type":
                field_value = cls.__name__
            else:
                field_value = field.get_default()

            if field_value is None:
                field_value = field.field_info.extra.get("example_value", None)

            is_enum = False
            if isinstance(field_value, Enum):
                field_value = field_value.value
                is_enum = True

            definition = getattr(field.type_, "__name__", None)
            if definition and (definition in schema_dict.definitions) and (not is_enum):
                field_value = field.type_.example(False)

            data[field.alias] = field_value

        if in_list:
            data = [data]

        return data

    @classmethod
    def example_json(cls, indent: int = 2) -> str:
        """Create a JSON example for an inputs model."""
        return json.dumps(cls.example(), indent=indent)


class PyDSSControllerModel(DiscoBaseModel):
    """PV Controller on deployment"""

    controller_type: ControllerType = Field(
        #alias="type",
        title="controller_type",
        description="The controller type defined in PyDSS.",
        example_value="PvController"
    )
    name: str = Field(
        title="name",
        description="The name of the controller",
        max_length=120,
        example_value="ctrl-name"
    )
    targets: Optional[Union[FilePath, List[FilePath]]] = Field(
        title="targets",
        default=None,
        description="The PV system files that need to apply controller."
    )

    class Config:
        title = "PyDSSControllerModel"
        anystr_strip_whitespace = True
        validate_assignment = True

    @root_validator(pre=True)
    def validate_pydss_controller_registration(cls, values: dict) -> dict:
        controller_type = ControllerType(values["controller_type"])
        name = values["name"]

        if name in registered_pydss_controllers[controller_type]:
            return values

        pydss_registry = Registry()
        registered = pydss_registry.is_controller_registered(
            controller_type=controller_type.value,
            name=name
        )
        if (name not in DISCO_CONTROLLER_NAMES) and not registered:
            raise ValueError("Invalid controller name.")

        registered_pydss_controllers[controller_type].add(name)
        return values


class OpenDssDeploymentModel(DiscoBaseModel):
    """PV Deployment on Feeder"""

    is_standalone: bool = Field(
        title="is_standalone",
        description="Set to True if the models are not to be modified by DISCO.",
        default=False,
    )
    deployment_file: str = Field(
        title="deployment_file",
        description="The path to the PV deployment file.",
        max_length=150,
        example_value="model-inputs/J1/PVDeployments/deployment_001.dss",
    )
    substation: str = Field(
        default=None,
        title="substation",
        description="The substation name.",
        max_length=120,
        example_value="substation_name",
    )
    feeder: str = Field(
        title="feeder",
        description="The feeder name in string.",
        max_length=120,
        example_value="J1",
    )
    dc_ac_ratio: float = Field(
        default=None,
        title="dc_ac_ratio",
        description=(
            "Factor representing the ratio between the PV plant's kW "
            "dc rating and its inverter's kW ac rating"
        ),
        gt=0.0,
        example_value=1.15,
    )
    directory: Optional[DirectoryPath] = Field(
        title="directory",
        default="model-inputs",
        description="The directory of OpenDSS models.",
    )
    kva_to_kw_rating: float = Field(
        default=None,
        title="kva_to_kw_rating",
        description=(
            "Factor representing the ratio between the inverter's input kW "
            "and output kVA ratings"
        ),
        gt=0.0,
        example_value=1.0,
    )
    project_data: dict = Field(
        title="project_data",
        default={},
        description=(
            "Opaque data, unused by DISCO, that will be forwarded through "
            "the stack for use in analysis/post-processing scripts."
        )
    )
    pydss_controllers: Union[PyDSSControllerModel, List[PyDSSControllerModel]] = Field(
        title="pydss_controllers",
        default=None,
        description="One or multiple controllers"
    )

    class Config:
        title = "OpenDssDeploymentModel"
        anystr_strip_whitespace = True
        validate_assignment = True


class SimulationModel(DiscoBaseModel):
    """DISCO Simulation Configurations"""

    start_time: datetime = Field(
        title="start_time",
        description="The start datetime of simulation.",
        example_value="2014-06-17T15:00:00.000",
    )
    end_time: datetime = Field(
        title="end_time",
        description="The end datetime of simulation.",
        example_value="2014-06-17T15:00:00.000",
    )
    step_resolution: int = Field(
        title="step_resolution",
        description="The step resolution of simulation in seconds.",
        ge=1,
        default=900
    )
    simulation_type: SimulationType = Field(
        title="simulation_type",
        default=SimulationType.SNAPSHOT,
        description="The simulation type supported in DISCO."
    )

    class Config:
        title = "SimulationModel"
        validate_assignment = True

    @validator("start_time", pre=True)
    def validate_start_time(cls, value: Union[str, datetime]) -> str:
        try:
            return standardize_timestamp(value)
        except ValueError:
            raise

    @validator("end_time", pre=True)
    def validate_end_time(cls, value: Union[str, datetime]) -> str:
        try:
            return standardize_timestamp(value)
        except ValueError:
            raise

    @root_validator(pre=True)
    def check_start_end_time(cls, values):
        """Validate start/end time for Snapshot simulation."""
        simulation_type = SimulationType(values.get("simulation_type"))
        is_snapshot = simulation_type == SimulationType.SNAPSHOT
        if not is_snapshot:
            return values

        start_time, end_time = values.get("start_time"), values.get("end_time")
        if is_snapshot and start_time != end_time:
            raise ValueError(
                "Invalid 'start_time' and 'end_time' values on "
                "Snapshot simulation, they should be same."
            )
        return values


class BaseAnalysisModel(DiscoBaseModel):
    """Base input model for DISCO simulation types."""
    model_type: Optional[str] = Field(
        title="model_type",
        description="model type.",
        max_length=255,
        default="BaseAnalysisModel"
    )
    name: str = Field(
        title="name",
        description="A unique name for this simulation job.",
        max_length=255,
        example_value="J1_123_Sim_456",
    )
    model_type: Optional[str] = Field(
        title="model_type",
        description="model type.",
        max_length=255
    )
    blocked_by: Optional[Set[str]] = Field(
        title="blocked_by",
        default=set(),
        description="A set of simulation job names that blocks currrent job.",
    )
    job_order: Optional[Union[int, float]] = Field(
        title="job_order",
        default=None,
        ge=0.0,
        description="The execution order of the simulation job.",
    )


class ImpactAnalysisBaseModel(BaseAnalysisModel, abc.ABC):
    """A base model for impact analysis types."""

    base_case: Optional[str] = Field(
        title="base_case",
        description="The base simulation job which has no added PV.",
        max_length=255,
    )
    is_base_case: Optional[bool] = Field(
        title="is_base_case",
        description="Whether this job is a base case",
        default=False,
    )
    deployment: OpenDssDeploymentModel = Field(
        title="deployment",
        description="PV deployment on feeder",
    )
    simulation: SimulationModel = Field(
        title="simulation",
        description="Simulation parameters with PV deployment",
    )

    class Config:
        title = "ImpactAnalysisBaseModel"
        anystr_strip_whitespace = True
        validate_assignment = True



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 24 15:26:33 2022

@author: senam
"""

import logging
import os
import sys
from pathlib import Path

import click
import opendssdirect as dss

from jade.loggers import setup_logging


logger = logging.getLogger("swad")


def get_load_info():
    """
    This function returns the following information pertaining to loads on a feeder:
        load name
        bus where each load is connected
        kW load capacity
        number of phases
    """
    load_info = {}
    # load_names = dss.Loads.AllNames()

    dss.Circuit.SetActiveClass("Load")
    flag = dss.ActiveClass.First()
    while flag > 0:
        name = dss.CktElement.Name()
        bus = dss.CktElement.BusNames()[0]
        load_kw = sum(dss.CktElement.Powers()[0::2])
        nphases = dss.CktElement.NumPhases()
        load_info[name] = {"name": name, "bus": bus, "kW": load_kw, "nphases": nphases}
        flag = dss.ActiveClass.Next()

    return load_info


def get_existing_pvs():
    total_existing_pv = 0
    existing_pv_dict = {}
    existing_pv_dist_dict = {}
    flag = dss.PVsystems.First()

    while flag > 0:
        bus = dss.Properties.Value("bus1")
        if bus not in existing_pv_dict:
            dss.Circuit.SetActiveBus(bus)
            existing_pv_dist_dict[bus] = dss.Bus.Distance()
            existing_pv_dict[bus] = dss.PVsystems.kVARated()
        else:
            existing_pv_dict[bus] += dss.PVsystems.kVARated()

        total_existing_pv += dss.PVsystems.kVARated()
        flag = dss.PVsystems.Next()

    return total_existing_pv, existing_pv_dict, existing_pv_dist_dict


def compute_weighted_average_pv_distance(load_info, option="size"):
    wavg_pv_distance = 0
    total_existing_pv, existing_pv_dict, existing_pv_dist_dict = get_existing_pvs()

    slack_dico = {k: v * existing_pv_dist_dict[k] for k, v in existing_pv_dict.items()}

    if option == "size":
        wavg_pv_distance = sum(slack_dico.values()) / max(0.001, total_existing_pv)

    elif option == "nodal_ratio":
        load_per_bus = {v["bus"]: v["kW"] for k, v in load_info.items()}

        slack_dico = {
            k: v * existing_pv_dist_dict[k] / max(0.001, load_per_bus[k])
            for k, v in existing_pv_dict.items()
        }
        ratio_sum = sum([v / max(0.001, load_per_bus[k]) for k, v in existing_pv_dict.items()])
        wavg_pv_distance = sum(slack_dico.values()) / max(0.001, ratio_sum)

    return wavg_pv_distance


def collect_weighted_average_pv_distance(filename: Path, option):
    orig = os.getcwd()
    try:
        reply = dss.Text.Command(f"compile '{filename}'")
        if reply is not None:
            logger.error("Failed to compile OpenDSS model %s: %s", filename, reply)
            sys.exit(1)

        load_info = get_load_info()
        avg = compute_weighted_average_pv_distance(load_info, option=option)
        logger.info("Computed weighted average PV distance %s for %s", avg, filename)
    finally:
        os.chdir(orig)
    return avg


def collect_weighted_average_pv_distances(path: Path, option):
    results = []
    for substation_path in path.iterdir():
        if substation_path.is_dir():
            for feeder_path in substation_path.iterdir():
                pv_path = feeder_path / "PVDeployments"
                if pv_path.is_dir():
                    for filename in pv_path.iterdir():
                        job_name = filename.name.replace(".dss", "")
                        fields = job_name.split("__")
                        if len(fields) != 5:
                            continue
                        substation, feeder, placement, sample, penetration_level = fields
                        wavg_pv_distance = collect_weighted_average_pv_distance(filename, option)
                        results.append(
                            {
                                "job_name": job_name,
                                "substation": substation,
                                "feeder": feeder,
                                "placement": placement,
                                "sample": sample,
                                "penetration_level": penetration_level,
                                "weighted_average_pv_distance": wavg_pv_distance,
                                "option": option,
                            }
                        )
    return results


@click.command()
@click.argument("path", type=click.Path(exists=True), callback=lambda _, __, x: Path(x))
@click.option(
    "--option",
    default="size",
    type=click.Choice(["size", "nodal_ratio"], case_sensitive=False),
    show_default=True,
    help="Computation methodology",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def compute_pv_swad(path, option, verbose):
    """Compute weighted average PV distance for each model. Requires an EnergyMeter."""
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging("swad", None, console_level=level, packages=["disco"])
    results = collect_weighted_average_pv_distances(path, option)
    if results:
        output_file = path / "weighted_average_pv_distances.csv"
        fields = results[0].keys()
        with open(output_file, "w") as f_out:
            f_out.write(",".join(fields))
            f_out.write("\n")
            for result in results:
                f_out.write(",".join([str(x) for x in result.values()]))
                f_out.write("\n")
        logger.info("Created %s", output_file)
    else:
        logger.info("There were no PV deployment files in %s", path)


if __name__ == "__main__":
    compute_pv_swad()



"""Exceptions raised by DISCO"""

from jade.exceptions import JadeBaseException


class DiscoBaseException(JadeBaseException):
    """All DISCO exceptions should derive from this."""


class AnalysisRunException(DiscoBaseException):
    """Raise when Analysis post-process run failed."""


class AnalysisConfigurationException(DiscoBaseException):
    """Raise when the configuration is not valid."""


class ExceededParallelLinesLimit(DiscoBaseException):
    """Raise when an upgrade simulation exceeds the limit for parallel lines"""


class ExceededParallelTransformersLimit(DiscoBaseException):
    """Raise when an upgrade simulation exceeds the limit for parallel transformers"""


class UpgradesInvalidViolationIncrease(DiscoBaseException):
    """Raise when violations increase unexpectedly"""


class OpenDssCompileError(DiscoBaseException):
    """Raise when OpenDSS fails to compile a model"""


class OpenDssConvergenceError(DiscoBaseException):
    """Raise when OpenDSS fails to converge"""


class PyDssConvergenceError(DiscoBaseException):
    """Raise when PyDSS fails to converge"""


class PyDssConvergenceErrorCountExceeded(DiscoBaseException):
    """Raise when PyDSS exceeds its threshold for error counts"""


class PyDssConvergenceMaxError(DiscoBaseException):
    """Raise when PyDSS exceeds its max tolerance error threshold"""


class UpgradesExternalCatalogRequired(DiscoBaseException):
    """Raise when an upgrade simulation needs an external catalog"""


class UpgradesExternalCatalogMissingObjectDefinition(DiscoBaseException):
    """Raise when an upgrade external catalog is missing an equipment type"""


class InvalidOpenDssElementError(DiscoBaseException):
    """Raise when an OpenDSS element has unexpected properties"""


EXCEPTIONS_TO_ERROR_CODES = {
    
    AnalysisConfigurationException: {
        "description": "The input configuration is invalid.",
        "error_code": 114,
    },
    ExceededParallelLinesLimit: {
        "description": "An Upgrades simulation exceeded the limit for parallel lines.",
        "corrective_action": "If not already done, enable ThermalUpgradeParamsModel.read_external_catalog"
        "and provide an external catalog. Or, increase ThermalUpgradeParamsModel.parallel_lines_limit to allow more parallel equipment to"
        "be placed to resolve thermal violations.",
        "error_code": 115,
    },
    ExceededParallelTransformersLimit: {
        "description": "An Upgrades simulation exceeded the limit for parallel transformers.",
        "corrective_action": "If not already done, enable ThermalUpgradeParamsModel.read_external_catalog"
        "and provide an external catalog. Or, increase ThermalUpgradeParamsModel.parallel_transformers_limit to allow more parallel equipment to"
        "be placed to resolve thermal violations.",
        "error_code": 116,
    },
    InvalidOpenDssElementError: {
        "description": "An OpenDSS element has unexpected properties.",
        "corrective_action": "Check the error message and fix the OpenDSS element definitions.",
        "error_code": 117,
    },
    OpenDssCompileError: {
        "description": "OpenDSS failed to compile a model.",
        "corrective_action": "Check the error message and fix the OpenDSS model definitions.",
        "error_code": 118,
    },
    OpenDssConvergenceError: {
        "description": "OpenDSS failed to converge.",
        "corrective_action": "Check the OpenDSS model. Also, refer to the OpenDSS manual to vary settings for convergence.",
        "error_code": 119,
    },
    PyDssConvergenceError: {
        "description": "PyDSS failed to find a solution in its external controls.",
        "error_code": 120,
    },
    PyDssConvergenceErrorCountExceeded: {
        "description": "PyDSS external controls exceeded the threshold for error counts.",
        "error_code": 121,
    },
    PyDssConvergenceMaxError: {
        "description": "PyDSS external controls exceeded the max tolerance error threshold.",
        "error_code": 122,
    },
    UpgradesExternalCatalogRequired: {
        "description": "An Upgrades simulation requires an external catalog for thermal upgrades in order to add or "
        "upgrade a component.",
        "corrective_action": "Provide an external catalog or disable ThermalUpgradeParamsModel.read_external_catalog.",
        "error_code": 123,
    },
    UpgradesExternalCatalogMissingObjectDefinition: {
        "description": "The Upgrades external catalog does not define a required object.",
        "corrective_action": "Ensure the external catalog defines all required objects. Refer to "
        "the error message for specific details.",
        "error_code": 124,
    },
    UpgradesInvalidViolationIncrease: {
        "description": "An Upgrades simulation detected an invalid increase in violations.",
        "corrective_action": "This could happen in cases when lines or transformers are extremely overloaded. Check and "
        "modify OpenDSS model for such instances.",
        "error_code": 125,
    },
}


def get_error_code_from_exception(exception_class):
    """Return the error code for a disco exception."""
    if not issubclass(exception_class, DiscoBaseException):
        raise Exception(f"exception={exception_class} is not handled")

    if exception_class in EXCEPTIONS_TO_ERROR_CODES:
        return EXCEPTIONS_TO_ERROR_CODES[exception_class]["error_code"]

    # Return a generic error code for any exceptions that we don't specifically want to map.
    return 1


def is_convergence_error(error_code):
    """Return True if the error code indicates a convergence error."""
    return error_code in {119, 120, 121, 122}



"""Defines DistributionInputs."""

import enum
import logging
import os

from jade.exceptions import InvalidParameter
from jade.jobs.job_inputs_interface import JobInputsInterface
from jade.utils.utils import load_data
from disco.distribution.deployment_parameters import DeploymentParameters


logger = logging.getLogger(__name__)


class DistributionInputs(JobInputsInterface):
    """Helper class to access Distribution input files."""

    _CONFIG_FILE = "configurations.json"

    def __init__(self, base_directory):
        """Constructs DistributionInputs.

        Parameters
        ----------
        base_directory : str

        """
        self._base = base_directory
        self._parameters = {}

        if not os.path.exists(self._base):
            raise InvalidParameter("inputs directory does not exist: {}"
                                   .format(self._base))

        self._parse_config_files()

    def __repr__(self):
        text = [str(x) for x in self.list_keys()]
        return "\n".join(text)

    def _parse_config_files(self):
        filename = os.path.join(self._base, self._CONFIG_FILE)
        data = load_data(filename)
        for job_data in data:
            job = DeploymentParameters(**job_data)
            assert job.name not in self._parameters
            self._parameters[job.name] = job

    @property
    def base_directory(self):
        return self._base

    def get_job(self, key):
        """Get the job object from a key.

        Parameters
        ----------
        key : namedtuple

        Returns
        -------
        DeploymentParameters

        Raises
        ------
        InvalidParameter
            thrown if the key is not stored

        """
        job = self._parameters.get(key)
        if job is None:
            raise InvalidParameter(f"invalid key {key}")

        return job

    def get_available_parameters(self):
        return self._parameters

    def _get_unique_parameters(self, param):
        assert param != "deployment"
        params = set()
        for item in self._parameters.values():
            val = getattr(item, param)
            if isinstance(val, enum.Enum):
                val = val.value
            params.add(val)

        params = list(params)
        params.sort()
        return params

    def iter_jobs(self):
        """Return an iterator of jobs."""
        return self._parameters.values()

    def list_jobs(self):
        """List the available jobs.

        Returns
        -------
        list
            list of DeploymentParameters

        """
        return list(self._parameters.values())

    def list_feeders(self):
        """List available feeders.

        Returns
        -------
        list
            list of strings

        """
        return self._get_unique_parameters("feeder")

    def list_parameters(self, feeders=None, dc_ac_ratios=None,
                        kva_to_kw_ratings=None, deployment_names=None):
        """List the parameters after applying filters. The filter values can be
        single objects or lists.

        Returns
        -------
        list
            list of DeploymentParameters

        """
        if feeders is not None and not isinstance(feeders, list):
            feeders = [feeders]
        if dc_ac_ratios is not None and not isinstance(dc_ac_ratios, list):
            dc_ac_ratios = [dc_ac_ratios]
        if kva_to_kw_ratings is not None and not isinstance(kva_to_kw_ratings,
                                                            list):
            kva_to_kw_ratings = [kva_to_kw_ratings]
        if deployment_names is not None and not isinstance(deployment_names,
                                                           list):
            deployment_names = [deployment_names]

        parameters = []
        for param in self._parameters.values():
            if feeders is not None and param.feeder not in feeders:
                continue
            if dc_ac_ratios is not None and \
                    param.deployment.dc_ac_ratio not in dc_ac_ratios:
                continue
            if kva_to_kw_ratings is not None and \
                    param.deployment.kva_to_kw_rating not in kva_to_kw_ratings:
                continue
            if deployment_names is not None and \
                    param.deployment.name not in deployment_names:
                continue
            parameters.append(param)

        return parameters

    def list_keys(self):
        """List the available job keys.

        Returns
        -------
        list
            list of namedtuple

        """
        keys = list(self._parameters.keys())
        keys.sort()
        return keys




from jade.exceptions import InvalidParameter
from disco.enums import AnalysisModelType
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.models.time_series_analysis_model import TimeSeriesAnalysisModel
from disco.models.upgrade_cost_analysis_model import UpgradeCostAnalysisModel

_MAPPING = {
    AnalysisModelType.SnapshotImpactAnalysis: SnapshotImpactAnalysisModel,
    AnalysisModelType.TimeSeriesImpactAnalysis: TimeSeriesAnalysisModel,
    AnalysisModelType.UpgradeCostAnalysis: UpgradeCostAnalysisModel
}

_MAPPING_BY_NAME = {
    "SnapshotImpactAnalysisModel": SnapshotImpactAnalysisModel,
    "TimeSeriesAnalysisModel": TimeSeriesAnalysisModel,
    "UpgradeCostAnalysisModel": UpgradeCostAnalysisModel
}


def get_model_class_by_analysis_type(analysis_type):
    """Return the model class for an analysis_type.

    Parameters
    ----------
    analysis_type : AnalysisModelType

    Returns
    -------
    class
        Will be a subtype of DiscoBaseModel

    """
    if analysis_type not in _MAPPING:
        raise InvalidParameter(f"no mapping for {analysis_type}")
    return _MAPPING[analysis_type]


def get_model_class_by_name(name):
    """Return the model class for a name.

    Parameters
    ----------
    name : str

    Returns
    -------
    class
        Will be a subtype of DiscoBaseModel

    """
    if name not in _MAPPING_BY_NAME:
        raise InvalidParameter(f"no mapping for {name}")
    return _MAPPING_BY_NAME[name]


def make_model(data):
    """Make a model from serialized data.

    Parameters
    ----------
    data : dict

    Returns
    -------
    class
        Will be a subtype of DiscoBaseModel

    """
    model_class = get_model_class_by_name(data["model_type"])
    return model_class.validate(data)


def list_model_classes():
    """Return all simulation input models.

    Returns
    -------
    list
        list of model classes

    """
    return [
        SnapshotImpactAnalysisModel,
        TimeSeriesAnalysisModel
    ]



from typing import List, Optional, Set, Dict
from pydantic.v1 import Field, root_validator, validator
from pydantic.v1.error_wrappers import ValidationError

import pandas as pd

from PyDSS.controllers import PvControllerModel

from disco.models.base import BaseAnalysisModel
from disco.models.upgrade_cost_analysis_equipment_model import *
from disco.extensions.upgrade_simulation.upgrade_configuration import DEFAULT_UPGRADE_PARAMS_FILE

_DEFAULT_UPGRADE_PARAMS = None
_SUPPORTED_UPGRADE_TYPES = ["thermal", "voltage"]


def _get_default_upgrade_params():
    global _DEFAULT_UPGRADE_PARAMS
    if _DEFAULT_UPGRADE_PARAMS is None:
        _DEFAULT_UPGRADE_PARAMS = load_data(DEFAULT_UPGRADE_PARAMS_FILE)
    return _DEFAULT_UPGRADE_PARAMS


def _extract_specific_model_properties_(model_name, field_type_key, field_type_value):    
    return [field_name  for field_name in model_name.schema().get("properties") if model_name.schema()["properties"][field_name].get(field_type_key) == field_type_value]


def get_default_thermal_upgrade_params():
    return _get_default_upgrade_params()["thermal_upgrade_params"]


def get_default_voltage_upgrade_params():
    return _get_default_upgrade_params()["voltage_upgrade_params"]


class OpenDSSLineModel(OpenDSSLineParams):
    bus1: str = Field(
        title="bus1",
        description="bus1",
    )
    bus2: str = Field(
        title="bus2",
        description="bus2",
    )
    length: float = Field(
        title="length",
        description="length",
    )
    enabled: Any = Field(
        title="enabled",
        description="enabled",
    )


class LineModel(OpenDSSLineModel, ExtraLineParams):
    name: str = Field(
        title="name",
        description="name. This is not a direct OpenDSS object property.",
    )


class LineCatalogModel(OpenDSSLineParams, ExtraLineParams):
    """Contains Line information needed for thermal upgrade analysis. Most fields can be directly obtained from the opendss models"""
    name: str = Field(
        title="name",
        description="name. This is not a direct OpenDSS object property.",
    )


class LineCodeCatalogModel(CommonLineParameters):
    """Contains LineCode information needed for thermal upgrade analysis. Most fields can be directly obtained from the opendss models"""
    name: str = Field(
        title="name",
        description="name",
    )
    nphases: int = Field(
        title="nphases",
        description="nphases",
        determine_upgrade_option=True,
    )
    Kron: Optional[Any] = Field(
        title="Kron",
        description="Kron",
        default="N",
        determine_upgrade_option=True,
    )
    neutral: float = Field(
        title="neutral",
        description="neutral",
        determine_upgrade_option=True,
    )
    like: Optional[Any] = Field(
        title="like",
        description="like",
        default=None,
    )
    baseFreq: float = Field(
        title="basefreq",
        description="basefreq",
        determine_upgrade_option=True,
    )
    

class LineGeometryCatalogModel(UpgradeParamsBaseModel):
    """Contains LineGeometry information needed for thermal upgrade analysis. Most fields can be directly obtained from the opendss models"""
    name: str = Field(
        title="name",
        description="name",
    )
    equipment_type: str = Field(
        title="equipment_type",
        description="equipment_type",
        determine_upgrade_option=True,
    )
    nconds: int = Field(
        title="nconds",
        description="nconds",
        determine_upgrade_option=True,
    )
    nphases: int = Field(
        title="nphases",
        description="nphases",
        determine_upgrade_option=True,
    )
    cond: int = Field(
        title="cond",
        description="cond",
        determine_upgrade_option=True,
    )
    wire: str = Field(
        title="wire",
        description="wire",
        determine_upgrade_option=True,
    )
    x: List[float] = Field(
        title="x",
        description="x",
        determine_upgrade_option=True,
    )
    h: List[float] = Field(
        title="h",
        description="h",
        determine_upgrade_option=True,
    )
    units: List[str] = Field(
        title="units",
        description="units",
        determine_upgrade_option=True,
    )
    normamps: float = Field(
        title="normamps",
        description="normamps",
        determine_upgrade_option=True,
    )
    emergamps: float = Field(
        title="emergamps",
        description="emergamps",
        determine_upgrade_option=True,
    )
    reduce: str = Field(
        title="reduce",
        description="reduce",
        determine_upgrade_option=True,
    )
    spacing: Optional[Any] = Field(
        title="spacing",
        description="spacing",
    )
    wires: List[str] = Field(
        title="wires",
        description="wires",
        determine_upgrade_option=True,
    )
    cncable: str = Field(
        title="cncable",
        description="cncable",
        determine_upgrade_option=True,
    )
    tscable: str = Field(
        title="tscable",
        description="tscable",
        determine_upgrade_option=True,
    )
    cncables: List[str] = Field(
        title="cncables",
        description="cncables",
        determine_upgrade_option=True,
    )
    tscables: List[str] = Field(
        title="tscables",
        description="tscables",
        determine_upgrade_option=True,
    )
    Seasons: int = Field(
        title="Seasons",
        description="Seasons",
    )
    Ratings: Any = Field(
        title="Ratings",
        description="Ratings",
    )
    LineType: str = Field(
        title="LineType",
        description="LineType",
        determine_upgrade_option=True,
    )
    like: Any = Field(
        title="like",
        description="like",
    )
    LineType: str = Field(
        title="LineType",
        description="LineType",
        determine_upgrade_option=True,
    )


class OpenDSSTransformerModel(CommonTransformerParameters):
    bus: str = Field(
        title="bus",
        description="bus",
    )
    buses: Any = Field(
        title="buses",
        description="buses",
    )
    enabled: Any = Field(
        title="enabled",
        description="enabled",
    )


class TransformerModel(OpenDSSTransformerModel, ExtraTransformerParams):
    name: str = Field(
        title="name",
        description="name. This is not a direct OpenDSS object property.",
    )


class TransformerCatalogModel(CommonTransformerParameters, ExtraTransformerParams):
    """Contains Transformer information needed for thermal upgrade analysis. Most fields can be directly obtained from the opendss models"""
    name: str = Field(
        title="name",
        description="name",
    )


class UpgradeTechnicalCatalogModel(UpgradeParamsBaseModel):
    """Contains Upgrades Technical Catalog needed for thermal upgrade analysis"""
    line: Optional[List[LineCatalogModel]] = Field(
        title="line",
        description="line catalog",
        default=[]
    )
    transformer: Optional[List[TransformerCatalogModel]] = Field(
        title="transformer",
        description="transformer catalog",
        default=[]
    )
    linecode: Optional[List[LineCodeCatalogModel]] = Field(
        title="linecode",
        description="linecode catalog",
        default=[]
    )
    geometry: Optional[List[LineGeometryCatalogModel]] = Field(
        title="linegeometry",
        description="linegeometry catalog",
        default=[]
    )


class ThermalUpgradeParamsModel(UpgradeParamsBaseModel):
    """Thermal Upgrade Parameters for all jobs in a simulation"""

    # Required fields
    transformer_upper_limit: float = Field(
        title="transformer_upper_limit",
        description="Transformer upper limit in per unit (example: 1.25)",
    )
    line_upper_limit: float = Field(
        title="line_upper_limit",
        description="Line upper limit in per unit (example: 1.25)",
    )
    line_design_pu: float = Field(
        title="line_design_pu",
        description="Line design in per unit (example: 0.75)",
    )
    transformer_design_pu: float = Field(
        title="transformer_design_pu",
        description="Transformer design in per unit (example: 0.75)",
    )
    voltage_upper_limit: float = Field(
        title="voltage_upper_limit",
        description="Voltage upper limit in per unit (example: 1.05)",
    )
    voltage_lower_limit: float = Field(
        title="voltage_lower_limit",
        description="Voltage lower limit in per unit (example: 0.95)",
    )
    read_external_catalog: bool = Field(
        title="read_external_catalog",
        description="Flag to determine whether external catalog is to be used (example: False)",
    )
    external_catalog: str = Field(
        title="external_catalog",
        description="Location to external upgrades technical catalog json file",
    )

    # Optional fields
    create_plots: Optional[bool] = Field(
        title="create_plots", description="Flag to enable or disable figure creation", default=True
    )
    parallel_transformers_limit: Optional[int] = Field(
        title="parallel_transformers_limit", description="Parallel transformer limit", default=4
    )
    parallel_lines_limit: Optional[int] = Field(
        title="parallel_lines_limit", description="Parallel lines limit", default=4
    )
    upgrade_iteration_threshold: Optional[int] = Field(
        title="upgrade_iteration_threshold", description="Upgrade iteration threshold", default=5
    )
    timepoint_multipliers: Optional[Dict] = Field(
        title="timepoint_multipliers",
        description='Dictionary to provide timepoint multipliers. example: timepoint_multipliers={"load_multipliers": {"with_pv": [1.2], "without_pv": [0.6]}}',
    )

    @validator("voltage_lower_limit")
    def check_voltage_lower_limits(cls, voltage_lower_limit, values):
        upper = values["voltage_upper_limit"]
        if upper <= voltage_lower_limit:
            raise ValueError(
                f"voltage_upper_limit={upper} must be greater than voltage_lower_limit={voltage_lower_limit}"
            )
        return voltage_lower_limit
    
    @validator("line_design_pu")
    def check_line_design_pu(cls, line_design_pu, values):
        upper = values["line_upper_limit"]
        if line_design_pu >= upper:
            raise ValueError(
                f"line_design_pu={line_design_pu} must be lesser than line_upper_limit={upper}"
            )
        return line_design_pu

    @validator("transformer_design_pu")
    def check_transformer_design_pu(cls, transformer_design_pu, values):
        upper = values["transformer_upper_limit"]
        if transformer_design_pu >= upper:
            raise ValueError(
                f"transformer_design_pu={transformer_design_pu} must be lesser than transformer_upper_limit={upper}"
            )
        return transformer_design_pu

    @validator("external_catalog")
    def check_catalog(cls, external_catalog, values):
        if values["read_external_catalog"]:
            if not Path(external_catalog).exists():
                raise ValueError(f"{external_catalog} does not exist")
            # Just verify that it constructs the model.
            UpgradeTechnicalCatalogModel(**load_data(external_catalog))
        return external_catalog

    @validator("timepoint_multipliers")
    def check_timepoint_multipliers(cls, timepoint_multipliers):
        if timepoint_multipliers is None:
            return timepoint_multipliers
        if "load_multipliers" not in timepoint_multipliers:
            raise ValueError("load_multipliers must be defined in timepoint_multipliers")
        if ("with_pv" not in timepoint_multipliers["load_multipliers"]) and ("without_pv" not in timepoint_multipliers["load_multipliers"]):
            raise ValueError(
                'Either "with_pv" or "without_pv" must be defined in timepoint_multipliers["load_multipliers"]'
            )
        return timepoint_multipliers


class VoltageUpgradeParamsModel(UpgradeParamsBaseModel):
    """Voltage Upgrade Parameters for all jobs in a simulation"""

    # Required fields
    initial_upper_limit: float = Field(
        title="initial_upper_limit",
        description="Initial upper limit in per unit (example: 1.05)",
    )
    initial_lower_limit: float = Field(
        title="initial_lower_limit",
        description="Initial lower limit in per unit (example: 0.95)",
    )
    final_upper_limit: float = Field(
        title="final_upper_limit",
        description="Final upper limit in per unit (example: 1.05)",
    )
    final_lower_limit: float = Field(
        title="final_lower_limit",
        description="Final lower limit in per unit (example: 0.95)",
    )
    nominal_voltage: float = Field(
        title="nominal_voltage",
        description="Nominal voltage (volts) (example: 120)",
    )

    # Optional fields
    create_plots: Optional[bool] = Field(
        title="create_plots", 
        description="Flag to enable or disable figure creation", 
        default=True
    )
    capacitor_sweep_voltage_gap: float = Field(
        title="capacitor_sweep_voltage_gap",
        description="Capacitor sweep voltage gap (example: 1)",
        default=1.0,
    )
    reg_control_bands: List[int] = Field(
        title="reg_control_bands",
        description="Regulator control bands (example: [1, 2])",
        default=[1, 2],
    )
    reg_v_delta: float = Field(
        title="reg_v_delta",
        description="Regulator voltage delta (example: 0.5)",
        default=0.5,
    )
    max_regulators: int = Field(
        title="max_regulators",
        description="Maximum number of regulators",
        default=4,
    )
    place_new_regulators: bool = Field(
        title="place_new_regulators",
        description="Flag to enable or disable new regulator placement",
        default=True,
    )
    use_ltc_placement: bool = Field(
        title="use_ltc_placement",
        description="Flag to enable or disable substation LTC upgrades module",
        default=True,
    )
    timepoint_multipliers: dict = Field(
        title="timepoint_multipliers",
        description='Dictionary to provide timepoint multipliers. example: timepoint_multipliers={"load_multipliers": {"with_pv": [1.2], "without_pv": [0.6]}}',
        default=None,
    )
    capacitor_action_flag: bool = Field(
        title="capacitor_action_flag",
        description="Flag to enable or disable capacitor controls settings sweep module",
        default=True,
    )
    existing_regulator_sweep_action: bool = Field(
        title="existing_regulator_sweep_action",
        description="Flag to enable or disable existing regulator controls settings sweep module",
        default=True,
    )
    max_control_iterations: int = Field(
        title="max_control_iterations",
        description="Max control iterations to be set for OpenDSS",
        default=50,
    )

    @validator("initial_lower_limit")
    def check_initial_voltage_lower_limits(cls, initial_lower_limit, values):
        upper = values["initial_upper_limit"]
        if upper <= initial_lower_limit:
            raise ValueError(
                f"initial_upper_limit={upper} must be greater than initial_lower_limit={initial_lower_limit}"
            )
        return initial_lower_limit
    
    @validator("final_lower_limit")
    def check_final_voltage_lower_limits(cls, final_lower_limit, values):
        upper = values["final_upper_limit"]
        if upper <= final_lower_limit:
            raise ValueError(
                f"final_upper_limit={upper} must be greater than final_lower_limit={final_lower_limit}"
            )
        return final_lower_limit
    
    @validator("timepoint_multipliers")
    def check_timepoint_multipliers(cls, timepoint_multipliers):
        if timepoint_multipliers is None:
            return timepoint_multipliers
        if "load_multipliers" not in timepoint_multipliers:
            raise ValueError("load_multipliers must be defined in timepoint_multipliers")
        if ("with_pv" not in timepoint_multipliers["load_multipliers"]) and ("without_pv" not in timepoint_multipliers["load_multipliers"]):
            raise ValueError(
                'Either "with_pv" or "without_pv" must be defined in timepoint_multipliers["load_multipliers"]'
            )
        return timepoint_multipliers


class UpgradeCostAnalysisGenericModel(BaseAnalysisModel):
    """Parameters for each job in a simulation"""

    name: str = Field(
        title="name",
        description="Unique name identifying the job",
    )
    opendss_model_file: str = Field(
        title="opendss_model_file",
        description="Path to file used load the simulation model files",
    )
    model_type: str = Field(
        title="model_type",
        description="Model type",
        default="UpgradeCostAnalysisGenericModel",
    )
    blocked_by: Set[str] = Field(
        title="blocked_by",
        description="Names of jobs that must finish before this job starts",
        default=set(),
    )
    estimated_run_minutes: Optional[int] = Field(
        title="estimated_run_minutes",
        description="Optionally advises the job execution manager on how long the job will run",
    )

    @validator("opendss_model_file")
    def check_model_file(cls, opendss_model_file):
        if not Path(opendss_model_file).exists():
            raise ValueError(f"{opendss_model_file} does not exist")
        return opendss_model_file


class PyDssControllerModels(UpgradeParamsBaseModel):
    """Defines the settings for PyDSS controllers"""

    pv_controller: Optional[PvControllerModel] = Field(
        title="pv_controller", description="Settings for a PV controller"
    )


class UpgradeCostAnalysisSimulationModel(UpgradeParamsBaseModel):
    """Defines the jobs in an upgrade cost analysis simulation."""

    class Config:
        title = "UpgradeCostAnalysisSimulationModel"
        anystr_strip_whitespace = True
        validate_assignment = True
        validate_all = True
        extra = "forbid"
        use_enum_values = False

    thermal_upgrade_params: ThermalUpgradeParamsModel = Field(
        default=ThermalUpgradeParamsModel(**get_default_thermal_upgrade_params())
    )
    voltage_upgrade_params: VoltageUpgradeParamsModel = Field(
        default=VoltageUpgradeParamsModel(**get_default_voltage_upgrade_params())
    )
    upgrade_cost_database: str = Field(
        title="upgrade_cost_database",
        description="Database containing costs for each equipment type",
    )
    calculate_costs: bool = Field(
        title="calculate_costs",
        description="If True, calculate upgrade costs from database.",
        default=True,
    )
    upgrade_order: List[str] = Field(
        description="Order of upgrade algorithm. 'thermal' or 'voltage' can be removed from the "
        "simulation by excluding them from this parameter.",
        default=_SUPPORTED_UPGRADE_TYPES,
    )
    pydss_controllers: PyDssControllerModels = Field(
        title="pydss_controllers",
        description="If enable_pydss_controllers is True, these PyDSS controllers are applied to each corresponding element type.",
        default=PyDssControllerModels(),
    )
    plot_violations: bool = Field(
        title="plot_violations",
        description="If True, create plots of violations before and after simulation.",
        default=True,
    )
    enable_pydss_controllers: bool = Field(
        title="enable_pydss_controllers",
        description="Flag to enable/disable use of PyDSS controllers",
        default=False,
    )
    include_pf1: bool = Field(
        title="include_pf1",
        description="Include PF1 scenario (no controls) if pydss_controllers are defined.",
        default=True,
    )
    dc_ac_ratio: Optional[float] = Field(
        title="dc_ac_ratio", description="Apply DC-AC ratio for PV Systems", default=None
    )
    jobs: List[UpgradeCostAnalysisGenericModel]

    @root_validator(pre=True)
    def check_job_names(cls, values):
        names = set()
        for job in values["jobs"]:
            if job["name"] in names:
                raise ValueError(f"{job['name']} is duplicated")
            names.add(job["name"])
        return values

    @validator("calculate_costs")
    def check_database(cls, calculate_costs, values):
        if calculate_costs:
            if "upgrade_cost_database" not in values:
                return calculate_costs
            if not Path(values["upgrade_cost_database"]).exists():
                raise ValueError(f"{values['upgrade_cost_database']} does not exist")
            # Just verify that it constructs the model.
            load_cost_database(values["upgrade_cost_database"])
        return calculate_costs

    @validator("upgrade_order")
    def check_upgrade_order(cls, upgrade_order):
        diff = set(upgrade_order).difference(_SUPPORTED_UPGRADE_TYPES)
        if diff:
            raise ValueError(f"Unsupported values in upgrade_order: {diff}")
        return upgrade_order

    def has_pydss_controllers(self):
        """Return True if a PyDSS controller is defined.

        Returns
        -------
        bool

        """
        return self.pydss_controllers.pv_controller is not None
 
    
class TransformerUnitCostModel(UpgradeParamsBaseModel):
    """Contains Transformer Unit Cost Database Model"""
    
    phases: int = Field(
        title="phases",
        description="Number of phases",
    )
    primary_kV: float = Field(
        title="primary_kV",
        description="Transformer primary winding voltage, in kV",
    )
    secondary_kV: float = Field(
        title="secondary_kV",
        description="Transformer secondary winding voltage, in kV",
    )    
    num_windings: int = Field(
        title="num_windings",
        description="Number of windings",
    )
    primary_connection_type: str = Field(
        title="primary_connection_type",
        description="Transformer primary winding connection type. Should be wye or delta",
    )
    secondary_connection_type: str = Field(
        title="secondary_connection_type",
        description="Transformer secondary winding connection type. Should be wye or delta",
    )
    rated_kVA: float = Field(
        title="rated_kVA",
        description="Transformer Rated kVA",
    )
    cost: float = Field(
        title="cost",
        description="Transformer unit cost",
    )
    cost_units: str = Field(
        title="cost_units",
        description="Unit for cost. This should be in USD/unit",
    )
    
    @validator("cost_units")    
    def check_transformer_cost_units(cls, cost_units):
        if cost_units not in ("USD/unit"):
            raise ValueError("Incorrect cost units")
        return cost_units
    
    @validator("primary_connection_type")
    def check_primary_connection(cls, primary_connection_type):
        if primary_connection_type not in ("wye", "delta"):
            raise ValueError("Incorrect transformer primary connection type")
        return primary_connection_type
    
    @validator("secondary_connection_type")
    def check_secondary_connection(cls, secondary_connection_type):
        if secondary_connection_type not in ("wye", "delta"):
            raise ValueError("Incorrect transformer secondary connection type")
        return secondary_connection_type
        

class LineUnitCostModel(UpgradeParamsBaseModel):
    """Contains Line Unit Cost Database Model"""
    
    upgrade_type: str = Field(
        title="upgrade_type",
        description="Description of whether this is a new_line or reconductored_line",
    )
    phases: int = Field(
        title="phases",
        description="Number of phases",
    )
    voltage_kV: float = Field(
        title="voltage_kV",
        description="Voltage level in kV",
    )    
    ampere_rating: float = Field(
        title="ampere_rating",
        description="Line rating in amperes",
    )
    line_placement: str = Field(
        title="line_placement",
        description="Placement of line: overhead or underground",
    )
    cost_per_m: float = Field(
        title="cost_per_m",
        description="Cost per meter",
    )
    cost_units: str = Field(
        title="cost_units",
        description="Unit for cost. This should be in USD",
    )
    name: Optional[str] = Field(
        title="name",
        description="Line name. This is an optional parameter, and is not used.",
    )
    
    @validator("line_placement")
    def check_line_placement(cls, line_placement):
        if line_placement not in ("underground", "overhead"):
            raise ValueError("Incorrect Line placement type. Acceptable values: overhead, underground.")
        return line_placement
    
    @validator("cost_units")
    def check_line_cost_units(cls, cost_units):
        if cost_units not in ("USD"):
            raise ValueError("Incorrect cost units")
        return cost_units
    
    @validator("upgrade_type")
    def check_line_upgrade_type(cls, upgrade_type):
        if upgrade_type not in ("new_line", "reconductored_line"):
            raise ValueError("Incorrect line upgrade_type")
        return upgrade_type
    
        
class ControlUnitCostModel(UpgradeParamsBaseModel):
    """Contains Control Changes Cost Database Model"""
    
    type: str = Field(
        title="type",
        description="Type of control setting",
    )
    cost: float = Field(
        title="cost",
        description="Control changes unit cost",
    )
    cost_units: str = Field(
        title="cost_units",
        description="Unit for cost. This should be in USD/unit",
    )
    
    @validator("cost_units")
    def check_control_cost_units(cls, cost_units):
        if cost_units not in ("USD/unit"):
            raise ValueError("Incorrect cost units")
        return cost_units
    
    
class VRegUnitCostModel(TransformerUnitCostModel):
    """Contains Voltage Regulator Cost Database Model"""
    
    type: str = Field(
        title="type",
        description="This should be 'Add new voltage regulator transformer'.",
    )
    

class MiscUnitCostModel(UpgradeParamsBaseModel):
    """Contains Miscellaneous Cost Database Model"""
    
    description: str = Field(
        title="description",
        description="Description of whether this is a fixed cost to add or replace transformer. "
        "These are optional, and will be used if provided.",
    )
    total_cost: float = Field(
        title="total_cost",
        description="total_cost",
    )
    cost_units: str = Field(
        title="cost_units",
        description="Unit for cost. This should be in USD/unit",
    )
    
    @validator("cost_units")
    def check_misc_cost_units(cls, cost_units):
        if cost_units not in ("USD/unit"):
            raise ValueError("Incorrect cost units")
        return cost_units
    
    @validator("description")
    def check_misc_description(cls, description):
        if description not in ("Replace transformer (fixed cost)", "Add new transformer (fixed cost)"):
            raise ValueError("Incorrect Miscellaneous Description")
        return description
    

class UpgradeCostDatabaseModel(UpgradeParamsBaseModel):
    """Contains Upgrades Unit Cost Database needed for cost analysis"""
    
    transformers: List[TransformerUnitCostModel] = Field(
        title="transformers",
        description="This consists of all transformer unit costs",
        default=[]
    )
    lines: List[LineUnitCostModel] = Field(
        title="lines",
        description="This consists of all line unit costs",
        default=[]
    )
    control_changes: List[ControlUnitCostModel] = Field(
        title="control_changes",
        description="This consists of all control changes unit costs",
        default=[]
    )
    voltage_regulators: List[VRegUnitCostModel] = Field(
        title="voltage_regulators",
        description="This consists of all voltage regulator unit costs",
        default=[]
    )
    misc: List[MiscUnitCostModel] = Field(
        title="misc",
        description="This consists of all miscellaneous unit costs",
        default=[]
    )


def load_cost_database(filepath):
    xfmr_cost_database = pd.read_excel(filepath, "transformers")
    line_cost_database = pd.read_excel(filepath, "lines")
    controls_cost_database = pd.read_excel(filepath, "control_changes")
    voltage_regulators_cost_database = pd.read_excel(filepath, "voltage_regulators")
    misc_database = pd.read_excel(filepath, "misc")

    transformers = xfmr_cost_database.to_dict(orient="records")
    lines = line_cost_database.to_dict(orient="records")
    control_changes = controls_cost_database.to_dict(orient="records")
    voltage_regulators = voltage_regulators_cost_database.to_dict(orient="records")
    misc = misc_database.to_dict(orient="records")

    # perform validation of input cost database using pydantic models
    # Check the first row of each type to ensure that the columns are correct.
    # If they aren't, pydantic will print errors for each row.
    for name, rows, cls in (
        ("transformers", transformers, TransformerUnitCostModel),
        ("lines", lines, LineUnitCostModel),
        ("control_changes", control_changes, ControlUnitCostModel),
        ("voltage_regulators", voltage_regulators, VRegUnitCostModel),
        ("misc", misc, MiscUnitCostModel),
    ):
        if rows:
            try:
                cls(**rows[0])
            except ValidationError:
                logger.error("Failed to validate cost database file=%s sheet=%s", filepath, name)
                raise

    UpgradeCostDatabaseModel(
        transformers=transformers,
        lines=lines,
        control_changes=control_changes,
        voltage_regulators=voltage_regulators,
        misc=misc,
    )
    return (
        xfmr_cost_database,
        line_cost_database,
        controls_cost_database,
        voltage_regulators_cost_database,
        misc_database,
    )




import logging
import os

import click

from jade.loggers import setup_logging
from jade.utils.utils import get_cli_string
from disco.analysis.postprocess_time_series import aggregate_deployments, compute_hosting_capacity


@click.command()
@click.argument("feeder")
@click.option(
    "-o", "--jade-runtime-output",
    required=True,
    help="jade runtime output directory",
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Enable debug logging",
)
def compute_time_series_impact_analysis(feeder, jade_runtime_output, verbose=False):
    """Run post-process computations for time series impact analysis."""
    level = logging.DEBUG if verbose else logging.INFO
    filename = os.path.join(jade_runtime_output, f"compute_time_series_impact_analysis_{feeder}.log")
    logger = setup_logging("disco", filename, console_level=level, file_level=level, packages=["disco"])
    logger.info("Run compute_time_series_impact_analysis for feeder %s", feeder)
    logger.info(get_cli_string())

    job_outputs = os.path.join(jade_runtime_output, "job-outputs")
    dfs = aggregate_deployments(job_outputs)



"""CLI to run a scenario."""
import logging

import opendssdirect as dss
from opendssdirect._version import __version__ as opendssdirect_version
from PyDSS import __version__ as pydss_version
from jade.jobs.job_configuration_factory import create_config_from_file
from disco.extensions.pydss_simulation.pydss_configuration import PyDssConfiguration
from disco.extensions.pydss_simulation.pydss_simulation import PyDssSimulation
from disco.version import __version__ as disco_version

logger = logging.getLogger(__name__)
DEFAULT_EXTENSION = "jade"


def auto_config(inputs, **kwargs):
    """
    Create a configuration file for PyDssSimulation.

    Parameters
    ----------
    inputs: str
        the path to directory containing PyDssSimulation data.

    """
    return PyDssConfiguration.auto_config(inputs, **kwargs)


def run(config_file, name, output, output_format, verbose):
    """Runs a PyDSS auto-generated scenario."""
    config = create_config_from_file(config_file, do_not_deserialize_jobs=True)
    job = config.get_job(name)

    logger.info("disco version = %s", disco_version)
    logger.info("PyDSS version = %s", pydss_version)
    logger.info("OpenDSSDirect version = %s", opendssdirect_version)
    logger.info("OpenDSS version = %s", dss.Basic.Version())

    simulation = PyDssSimulation.create(config.pydss_inputs,
                                        job,
                                        output=output)
    try:
        ret = simulation.run(verbose=verbose)
        return ret
    except Exception:
        logger.exception("unexpected exception in run pydss_simulation job=%s",
                         job.name)
        raise



import logging
import opendssdirect as dss

from PyDSS.simulation_input_models import ProjectModel
from PyDSS.common import SimulationType
from PyDSS.controllers import CircuitElementController, ControllerManager

from disco.exceptions import PyDssConvergenceError


logger = logging.getLogger(__name__)


def define_initial_pydss_settings(**kwargs):
    settings = ProjectModel(
        max_control_iterations=50,
        error_tolerance=0.0001,
        simulation_type=SimulationType.SNAPSHOT,
    )
    dss.Text.Command("Set ControlMode={}".format(settings.control_mode.value))
    dss.Solution.MaxControlIterations(settings.max_control_iterations)
    # we dont need to define controller everytime we solve the circuit, unless we're reloading the circuit
    controller = CircuitElementController(
        kwargs["pydss_volt_var_model"]
    )  # Use all elements.
    pydss_controller_manager = ControllerManager.create([controller], settings)
    kwargs.update({"pydss_controller_manager": pydss_controller_manager})
    return kwargs


def pydss_solve_and_check(raise_exception=False, **kwargs):
    logger.debug("Solving circuit using PyDSS controls")
    pydss_pass_flag = kwargs["pydss_controller_manager"].run_controls()
    if not pydss_pass_flag:
        logger.info(f"PyDSS Convergence Error")
        if raise_exception:
            raise PyDssConvergenceError("PyDSS solution did not converge")
    return pydss_pass_flag



"""Factory functions to create source models from source data"""

import os

from jade.exceptions import InvalidConfiguration
from jade.utils.utils import load_data
from disco.sources import EpriModel
from disco.sources import GemModel
from disco.sources import SourceTree1Model
from disco.sources import SourceTree2Model
from .base import FORMAT_FILENAME, TYPE_KEY

SUPPORTED_FORMATS = {
    "EpriModel": EpriModel,
    "GemModel": GemModel,
    "SourceTree1Model": SourceTree1Model,
    "SourceTree2Model": SourceTree2Model,
}


def list_subcommands(input_path):
    """List the available transform subcommands for the source model.

    Parameters
    ----------
    input_path : str
        filename or directory, depending on source data type

    Returns
    -------
    list

    """
    source_model = make_source_model(input_path)
    return source_model.list_transform_subcommands()


def make_source_model(input_path):
    """Construct the correct source model for input_path.

    Parameters
    ----------
    input_path : str
        filename or directory, depending on source data type

    Returns
    -------
    class
        derived from BaseSourceDataModel

    Raises
    ------
    InvalidConfiguration
        Raised if the format is not specified or supported.

    """
    if os.path.isdir(input_path):
        source_type = _make_source_model_from_directory(input_path)
    else:
        source_type = _make_source_model_from_file(input_path)

    if source_type not in SUPPORTED_FORMATS:
        raise InvalidConfiguration(f"{source_type} is not a supported format")

    return SUPPORTED_FORMATS[source_type]


def _make_source_model_from_directory(input_path):
    format_file = os.path.join(input_path, FORMAT_FILENAME)
    if not os.path.exists(format_file):
        raise InvalidConfiguration(f"{FORMAT_FILENAME} is not present in {input_path}")

    config = load_data(format_file)
    source_type = config.get(TYPE_KEY)
    if source_type is None:
        raise InvalidConfiguration(f"{TYPE_KEY} is not defined in {format_file}")

    return source_type


def _make_source_model_from_file(input_path):
    config = load_data(input_path)
    source_type = config.get(TYPE_KEY)
    if source_type is None:
        raise InvalidConfiguration(f"{TYPE_KEY} is not defined in {input_path}")

    return source_type



"""Base functionality for running PyDSS simulations."""

import abc
import copy
import logging
import os
import re

import PyDSS.exceptions as PyDssExceptions
from PyDSS.pydss_project import update_pydss_controllers
from PyDSS.pydss_project import PyDssProject, PyDssScenario

from jade.events import StructuredLogEvent, EVENT_CATEGORY_ERROR
from jade.jobs.job_execution_interface import JobExecutionInterface
from jade.loggers import log_event
from jade.utils.utils import dump_data
from jade.utils.timing_utils import timed_info

from disco.common import (
    EXIT_CODE_GOOD,
    LOADS_SUM_GROUP_FILENAME,
    PV_SYSTEMS_SUM_GROUP_FILENAME,
)
from disco.exceptions import (
    PyDssConvergenceError,
    PyDssConvergenceErrorCountExceeded,
    PyDssConvergenceMaxError,
)
from disco.pydss.common import ConfigType
from disco.events import EVENT_NO_CONVERGENCE
from disco.models.base import PyDSSControllerModel
from disco.pydss.pydss_utils import detect_convergence_problems


logger = logging.getLogger(__name__)


class PyDssSimulationBase(JobExecutionInterface, abc.ABC):
    """Runs a PyDss simulation."""

    _CONTROLLERS_FILENAME = "PvController.toml"
    _EXPORTS_FILENAME = "Exports.toml"
    _PYDSS_PROJECT_NAME = "pydss_project"
    _REGEX_YEAR = re.compile(r"^(\d\d\d\d)(-.*)")

    def __init__(self,
                 pydss_inputs,
                 deployment_params,
                 output,
                 # TODO: these two parameters should be config options
                 # and add_pct_pmpp should be required to be passed
                 add_pct_pmpp=True,
                 irradiance_scaling_factor=100):
        """Constructs a PyDssSimulation.

        Raises
        ------
        InvalidParameter
            Raised if any parameter is invalid.

        """
        self._pydss_inputs = copy.deepcopy(pydss_inputs)
        self._model = deployment_params.model
        self._output = output
        self._run_dir = os.path.join(
            self._output,
            self._model.name,
        )
        self._logs_dir = os.path.join(
            self._run_dir,
            "logs",
        )
        self._pydss_project = None
        self._dss_dir = None
        self._results_dir = None
        self._add_pct_pmpp = add_pct_pmpp
        self._irradiance_scaling_factor = irradiance_scaling_factor

    @property
    def model(self):
        return self._model

    def _get_control_mode(self):
        if self._model.model_type in (
            "SnapshotImpactAnalysisModel",
            "TimeSeriesAnalysisModel",
        ):
            return "Static"
        assert False, "unsupported type = {self._model.model_type}"

    @abc.abstractmethod
    def _get_deployment_input_path(self):
        """Return the full path to the user input deployment filename."""

    @abc.abstractmethod
    def _get_deployment_input_filename(self):
        """Return the user input deployment filename."""

    @abc.abstractmethod
    def _get_scenario_names(self):
        """Return the scenario name."""

    @staticmethod
    @abc.abstractmethod
    def _is_dss_file_path_absolute():
        """Return True if the deployment file path is absolute."""

    def _make_years_match(self, start_time_year, other_dict, other_dict_key):
        match = self._REGEX_YEAR.search(other_dict[other_dict_key])
        assert match, other_dict[other_dict_key]
        other_start_year = int(match.group(1))
        if start_time_year != other_start_year:
            logger.warning("start_time_year=%s does not match %s year=%s. Setting them to match.",
                           start_time_year, other_dict_key, other_start_year)
            other_dict[other_dict_key] = f"{start_time_year}{match.group(2)}"

    @abc.abstractmethod
    def _modify_open_dss_parameters(self):
        """Modify parameters in OpenDSS file; can be a no-op."""

    def _setup_pydss_project(self, verbose=False):
        dss_file_absolute_path = False
        if self._is_dss_file_path_absolute():
            deployment_filename = self._get_deployment_input_path()
            dss_file_absolute_path = True
        elif self._model.deployment.is_standalone:
            deployment_filename = self._model.deployment.deployment_file
            dss_file_absolute_path = True
        else:
            # PyDSS will join the directories with the name.
            deployment_filename = self._get_deployment_input_filename()

        dss_args = {
            "plots": {
                "create_dynamic_plots": False,
                "open_plots_in_browser": False,
            },
            "project": {
                "project_path": os.path.abspath(self._run_dir),
                "dss_file": deployment_filename,
                "dss_file_absolute_path": dss_file_absolute_path,
                "control_mode": self._get_control_mode(),
            },
            "logging": {
                "logging_level": "DEBUG" if verbose else "INFO",
                "enable_file": True,
                "enable_console": False,
                "clear_old_log_file": True,
                "log_time_step_updates": False,
            },
            "exports": {
                "hdf_max_chunk_bytes": 1048576,
            }
        }

        simulation_config = self._pydss_inputs[ConfigType.SIMULATION_CONFIG]
        self._modify_pydss_simulation_params(simulation_config["project"])

        for category, params in simulation_config.items():
            if category in dss_args:
                dss_args[category].update(params)
            else:
                dss_args[category] = params

        logger.info("PyDSS args: %s", dss_args)
        scenarios = self._make_pydss_scenarios()
        for scenario in scenarios:
            if scenario.snapshot_time_point_selection_config is not None:
                self._make_years_match(
                    self._model.simulation.start_time.year,
                    scenario.snapshot_time_point_selection_config,
                    "start_time",
                )
        self._pydss_project = PyDssProject.create_project(
            self._run_dir,
            self._PYDSS_PROJECT_NAME,
            scenarios,
            options=dss_args,
            master_dss_file=dss_args["project"]["dss_file"],
            force=True,  # This allows us to resubmit failed/missing jobs.
        )
        # TODO: this needs a better way of handling pf1
        self._apply_pydss_controllers(
            project_path=self._pydss_project.project_path,
            scenario_names=[s.name for s in scenarios if "pf1" not in s.name]
        )
        self._dss_dir = self._pydss_project.dss_files_path
        self._results_dir = [
            self._pydss_project.export_path(scenario_name)
            for scenario_name in self._get_scenario_names()
        ]

        logger.info("Setup folders at %s", self._run_dir)

        if not self._model.deployment.is_standalone:
            self._modify_open_dss_parameters()

    def _apply_pydss_controllers(self, project_path, scenario_names):
        """Update PyDSS controllers."""
        controllers = self._model.deployment.pydss_controllers
        if not controllers:
            return

        if isinstance(controllers, PyDSSControllerModel):
            controllers = [controllers]

        for controller in controllers:
            targets = controller.targets
            targets = targets or self._model.deployment.deployment_file
            if not isinstance(targets, list):
                targets = [targets]
            for scenario_name in scenario_names:
                for target in targets:
                    update_pydss_controllers(
                        project_path=project_path,
                        scenario=scenario_name,
                        controller_type=controller.controller_type.value,
                        controller=controller.name,
                        dss_file=target
                    )

    # TODO: not in use?
    def _get_pydss_scenarios(self):
        """Return a list of PyDssScenario objects."""
        return [
            PyDssScenario(x, exports=self._make_pydss_exports(x))
            for x in self._get_scenario_names()
        ]

    @abc.abstractmethod
    def _modify_pydss_simulation_params(self, config):
        pass

    def _make_pydss_exports(self, scenario):
        for class_name, properties in scenario["exports"].items():
            for vals in properties.values():
                if vals.get("sum_groups_file") == "assign at runtime":
                    if class_name == "Loads":
                        filename = LOADS_SUM_GROUP_FILENAME
                    elif class_name == "PVSystems":
                        filename = PV_SYSTEMS_SUM_GROUP_FILENAME
                    else:
                        raise Exception(f"Not supported: {class_name}")
                    vals["sum_groups_file"] = str(self._model.deployment.directory / "Metadata" / filename)
        filename = os.path.join(self._run_dir, self._EXPORTS_FILENAME)
        dump_data(scenario["exports"], filename)
        return filename

    def _make_pydss_scenarios(self):
        # TODO DT: how does this create different exports per scenario?
        return [
            PyDssScenario(
                x["name"],
                exports=self._make_pydss_exports(x),
                post_process_infos=x["post_process_infos"],
                snapshot_time_point_selection_config=x.get("snapshot_time_point_selection_config"),
            )
            for x in self._pydss_inputs[ConfigType.SCENARIOS]
        ]

    @property
    def pydss_project_path(self):
        if self._pydss_project is None:
            self._pydss_project = PyDssProject.load_project(
                os.path.join(self._run_dir, self._PYDSS_PROJECT_NAME)
            )
        return self._pydss_project.project_path

    @property
    def results_directory(self):
        return self._results_dir

    def list_results_files(self):
        results_files = {}
        scenario_names = self._get_scenario_names()
        for scenario_name, results_dir in zip(scenario_names, self._results_dir):
            if os.path.exists(results_dir):
                files = os.listdir(results_dir)
            else:
                # Exports may be disabled.
                files = []
            logger.debug("Results files: %s", files)
            results_files[scenario_name] = [os.path.join(results_dir, x) for x in files]
        return results_files

    def post_process(self, **kwargs):
        """Run post-processing on the output files."""

    @timed_info
    def run(self, verbose=False):
        """Runs the simulation."""
        self._setup_pydss_project()
        logger.info("Run simulation name=%s", self._model.name)
        logger.debug("Run simulation %s", self)

        orig_dir = os.getcwd()
        ret = EXIT_CODE_GOOD
        try:
            self._pydss_project.run(logging_configured=False, zip_project=True)
            self.check_convergence_problems()
        except PyDssExceptions.PyDssConvergenceError:
            logger.exception("Simulation failed with a convergence error")
            ret = PyDssConvergenceError
        except PyDssExceptions.PyDssConvergenceErrorCountExceeded:
            logger.exception("Simulation failed with a convergence error")
            ret = PyDssConvergenceMaxError
        except PyDssExceptions.PyDssConvergenceMaxError:
            logger.exception("Simulation failed with a convergence error")
            ret = PyDssConvergenceMaxError
        finally:
            os.chdir(orig_dir)

        # This may be used again in the future.
        # self.list_results_files()
        return ret

    def check_convergence_problems(self):
        """Logs events for convergence errors."""
        problems = detect_convergence_problems(self._pydss_project.project_path)
        if not problems:
            return

        # This is disabled because there can be huge counts. There is no need to duplicate them.
        #for problem in problems:
        #    event = StructuredLogEvent(
        #        source=self._model.name,
        #        category=EVENT_CATEGORY_ERROR,
        #        name=EVENT_NO_CONVERGENCE,
        #        message="Detected convergence problem in PyDSS log.",
        #        **problem,
        #    )
        #    log_event(event)

        logger.error("Job experienced with %s convergence problems", len(problems))



import os
import json
from pathlib import Path
import pickle
import math

import pandas as pd
import sklearn
import numpy as np

from disco.enums import SimulationHierarchy


# We observed that times were off by a little more than 2x.
FUDGE_FACTOR = 2.5


def generate_estimate_run_minutes(config):
    """Estimated run minutes.

    Parameters
    ----------
    config : PyDssConfiguration

    """
    def get_num_elem(dss_path,elem_name):
        n = 0
        with open(dss_path) as fp:
            for line in fp: 
                if "new "+elem_name.lower() in line.lower():
                    n += 1
        return n

    def compute_estimate(exe_time, scaling_factor):
        val = math.ceil(float(exe_time / 60) * FUDGE_FACTOR) * scaling_factor
        return max(val, 1)
    
    directory = os.path.dirname(__file__)
    model_file = os.path.join(directory, 'trained_lm_time_prediction.sav')
    trained_model = pickle.load(open(model_file, 'rb'))
    hierarchy = config.get_simulation_hierarchy()
    duration = None
    scale_factor = None
    
    for job in config.iter_jobs():
        if duration is None:
            seconds_per_day = 24 * 60 * 60
            duration = (job.model.simulation.end_time - job.model.simulation.start_time).total_seconds() / seconds_per_day
            scale_factor = duration / 365
        if hierarchy == SimulationHierarchy.SUBSTATION:
            deployment_dss = job.model.deployment.deployment_file
            sub_dir = job.model.deployment.directory
            path = Path(sub_dir)
            num_lines = 0
            for lines_dss in path.rglob("Lines.dss"):
                num_lines += get_num_elem(lines_dss,"Line")
            num_loads = 0
            for loads_dss in path.rglob("Loads.dss"):
                num_loads += get_num_elem(loads_dss,"Load")
            num_pvsystem = get_num_elem(deployment_dss,"pvsystem")
            exe_time_pred_s = trained_model.predict(np.array([num_lines, num_loads, num_pvsystem]).reshape(1,-1))
            job.estimated_run_minutes = compute_estimate(exe_time_pred_s, scale_factor)
        elif hierarchy == SimulationHierarchy.FEEDER:
            deployment_dss = job.model.deployment.deployment_file
            feeder_dir = job.model.deployment.directory
            lines_dss = os.path.join(feeder_dir, "OpenDSS","Lines.dss")
            loads_dss = os.path.join(feeder_dir, "OpenDSS","Loads.dss")
            num_lines = get_num_elem(lines_dss,"Line")
            num_loads = get_num_elem(loads_dss,"load")
            num_pvsystem = get_num_elem(deployment_dss,"pvsystem")
            exe_time_pred_s = trained_model.predict(np.array([num_lines, num_loads, num_pvsystem]).reshape(1,-1))
            job.estimated_run_minutes = compute_estimate(exe_time_pred_s, scale_factor)
        else:
            assert False, hierarchy


if __name__ == "__main__":
    generate_estimate_run_minutes()




import os
import shutil
import pytest


from tests.common import *


@pytest.fixture
def cleanup():
    def delete_files():
        for path in (
            PIPELINE_CONFIG,
            CONFIG_FILE,
            PRESCREEN_CONFIG_FILE,
            PRESCREEN_FINAL_CONFIG_FILE,
            TRANSFORM_MODEL_LOG,
            UPGRADE_SUMMARY
        ):
            if os.path.exists(path):
                os.remove(path)
        for path in (OUTPUT, MODELS_DIR):
            if os.path.exists(path):
                shutil.rmtree(path)

    delete_files()
    yield
    delete_files()


@pytest.fixture
def test_data_dir():
    """The path to the directory that contains the fixture data"""
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def smart_ds_substations():
    """The path to the tests data - smart-ds substations"""
    return os.path.join(
        os.path.dirname(__file__), "data", "smart-ds", "substations"
    )


@pytest.fixture
def smart_ds_substations_no_loadshapes():
    return os.path.join(
        os.path.dirname(__file__), "data", "smart-ds", "substations-no-load-shapes"
    )



"""Create summary files for hosting capacity results."""

import json
import logging
import re
import shutil
import sys
import time
from pathlib import Path
from tempfile import NamedTemporaryFile

import click
import chevron

from jade.loggers import setup_logging
from jade.utils.subprocess_manager import check_run_command
from jade.utils.utils import get_cli_string, load_data, dump_data

import disco


logger = logging.getLogger("summarize_hc_metrics")

DISCO = Path(disco.__path__[0]).parent
TEMPLATE_FILE = DISCO / "disco" / "postprocess" / "query.mustache"
HOSTING_CAPACITY_THRESHOLDS = DISCO / "disco" / "postprocess" / "config" / "hc_thresholds.toml"

SQLITE3_CMD_TEMPLATE = """
.read {{query_file}}
.headers on
.mode csv
.echo on
.output {{hc_summary_file}}
select * from hc_summary;
.output {{hc_by_sample_file}}
select * from hc_by_sample_kw;
.output {{bad_feeder_file}}
select * from bad_feeders order by feeder;
.output {{bad_feeder_pct_thresholds_file}}
select * from bad_feeders_pct_threshold order by feeder;
.output {{bad_feeder_violations_count_file}}
select * from bad_feeders_violation_count_overall;
"""


def _check_task_pattern(_, __, val):
    if val is None:
        return val
    if not re.search(r"^[\w\% ]+$", val):
        logger.error("Task pattern can only contain alphanumeric characters, spaces, and '%'.")
        sys.exit(1)
    return val


@click.command()
@click.option(
    "-d",
    "--database",
    type=click.Path(exists=True),
    required=True,
    help="Path to simulation results database file",
)
@click.option(
    "-s",
    "--scenario",
    type=click.Choice(["pf1", "control_mode", "derms"], case_sensitive=False),
    required=True,
    help="Scenario name",
)
@click.option(
    "-T",
    "--task-names",
    multiple=True,
    help="Query data with these task names in the database.",
)
@click.option(
    "-t",
    "--task-pattern",
    type=str,
    help="Pattern to match one or more tasks in the database with SQL LIKE. Can only contain "
    "letters, numbers, spaces, underscores, and %. Example: '%Time Series%",
    callback=_check_task_pattern,
)
@click.option(
    "--hc-thresholds",
    type=click.Path(),
    default=HOSTING_CAPACITY_THRESHOLDS,
    show_default=True,
    help="File containing thresholds for filtering metrics",
)
@click.option(
    "--thermal/--no-thermal",
    is_flag=True,
    default=True,
    show_default=True,
    help="Check for thermal violations",
)
@click.option(
    "--voltage/--no-voltage",
    is_flag=True,
    default=True,
    show_default=True,
    help="Check for voltage violations",
)
@click.option(
    "--secondaries/--no-secondaries",
    is_flag=True,
    default=False,
    show_default=True,
    help="Include secondary nodes in voltage checks.",
)
@click.option(
    "-o",
    "--output-directory",
    default="hc_reports",
    show_default=True,
    help="Create report files in this directory. Must not already exist.",
    callback=lambda _, __, x: Path(x),
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite any pre-existing output files.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose logging"
)
def summarize_hosting_capacity(
    database,
    scenario,
    task_names,
    task_pattern,
    hc_thresholds,
    thermal,
    voltage,
    secondaries,
    output_directory,
    force,
    verbose,
):
    """Create summary files for hosting capacity results."""
    if output_directory.exists():
        if force:
            shutil.rmtree(output_directory)
        else:
            print(f"{output_directory} already exists. Set --force to overwrite.", file=sys.stderr)
            sys.exit(1)
    output_directory.mkdir()

    level = logging.DEBUG if verbose else logging.INFO
    log_file = output_directory / "summarize_hc_metrics.log"
    setup_logging("summarize_hc_metrics", log_file, console_level=level, file_level=level, packages=["disco"])
    logger.info(get_cli_string())

    if task_pattern is None and not task_names:
        logger.error("One of --task-names or --tast-pattern must be passed")
        sys.exit(1)
    if task_pattern is not None and task_names:
        logger.error("Only one of --task-names and --tast-pattern can be passed")
        sys.exit(1)

    hc_summary_filename = output_directory / "hc_summary.csv"
    hc_by_sample_filename = output_directory / "hc_by_sample.csv"
    bad_feeder_filename = output_directory / "feeders_fail_base_case.csv"
    bad_feeder_pct_thresholds_filename = (
        output_directory / "feeders_pct_thresholds_fail_base_case.csv"
    )
    bad_feeder_violations_count_filename = (
        output_directory / "feeders_fail_base_case_threshold_violation_counts.csv"
    )

    defaults = load_data(hc_thresholds)
    options = {"scenario": scenario}
    if task_names:
        options["task_names"] = " OR ".join((f"task.name = '{x}'" for x in task_names))
    if task_pattern is not None:
        options["task_pattern"] = task_pattern
    if thermal:
        options["thermal"] = defaults["thermal"]
    if voltage:
        options["voltage"] = defaults["voltage"]
        options["voltage"]["secondaries"] = secondaries

    with open(TEMPLATE_FILE, "r") as f_in:
        query = chevron.render(f_in, options)

    with NamedTemporaryFile(mode="w") as f_query:
        f_query.write(query)
        f_query.flush()

        out_file = output_directory / "query.sql"
        shutil.copyfile(f_query.name, out_file)

        with NamedTemporaryFile(mode="w") as f_sqlite3_cmd:
            query_options = {
                "query_file": f_query.name,
                "hc_summary_file": hc_summary_filename,
                "hc_by_sample_file": hc_by_sample_filename,
                "bad_feeder_file": bad_feeder_filename,
                "bad_feeder_pct_thresholds_file": bad_feeder_pct_thresholds_filename,
                "bad_feeder_violations_count_file": bad_feeder_violations_count_filename,
            }
            f_sqlite3_cmd.write(chevron.render(SQLITE3_CMD_TEMPLATE, query_options))
            f_sqlite3_cmd.write("\n")
            f_sqlite3_cmd.flush()

            logger.info(
                "Running SQL queries on %s with thresholds\n%s",
                database,
                json.dumps(options, indent=4),
            )
            dump_data(options, output_directory / "thresholds.json", indent=True)
            start = time.time()
            cmd = f"sqlite3 -init {f_sqlite3_cmd.name} {database} .exit"
            check_run_command(cmd)
            logger.info("Queries complete. Duration = %.2f seconds", time.time() - start)

    for filename in output_directory.iterdir():
        logger.info("Created output file %s", filename)



"""
Defines EPRI Open DSS model.
The model can be transformed into DISCO OpenDSS model with PV deployments.
"""

import json
import logging
import os

import click

from jade.exceptions import InvalidParameter
from jade.utils.utils import ExtendedJSONEncoder
from PyDSS.common import ControllerType

from disco.cli.common import handle_existing_dir
from disco.enums import SimulationType, SimulationHierarchy
from disco.models.base import PyDSSControllerModel
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.models.time_series_analysis_model import TimeSeriesAnalysisModel
from disco.sources.base import (
    BaseOpenDssModel,
    SOURCE_CONFIGURATION_FILENAME,
    DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS,
    DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS,
)


logger = logging.getLogger(__name__)


COMMON_OPTIONS = (
    click.option(
        "-f",
        "--feeders",
        default=["all"],
        multiple=True,
        show_default=True,
        help="feeders to add; use 'all' to auto-detect and add all feeders",
    ),
    click.option(
        "-F",
        "--force",
        help="overwrite existing directory",
        is_flag=True,
        default=False,
        show_default=True
    )
)

def common_options(func):
    for option in reversed(COMMON_OPTIONS):
        func = option(func)
    return func


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-o",
    "--output",
    default=DEFAULT_SNAPSHOT_IMPACT_ANALYSIS_PARAMS["output_dir"],
    show_default=True,
    help="output directory",
)
@click.pass_context
def snapshot(ctx, feeders, force, start, output):
    """Transform input data for a snapshot simulation"""
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": start,
        "step_resolution": 900,
        "simulation_type": SimulationType.SNAPSHOT,
    }
    EpriModel.transform(
        input_path=input_path,
        output_path=output,
        simulation_params=simulation_params,
        simulation_model=SnapshotImpactAnalysisModel,
        feeders=feeders,
    )
    print(f"Transformed data from {input_path} to {output} for Snapshot Analysis.")


@click.command()
@common_options
@click.option(
    "-s",
    "--start",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["start_time"],
    show_default=True,
    help="simulation start time",
)
@click.option(
    "-e",
    "--end",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["end_time"],
    show_default=True,
    help="simulation end time",
)
@click.option(
    "-p",
    "--pv-profile",
    type=str,
    help="load shape profile name to apply to all PVSystems",
)
@click.option(
    "-r",
    "--resolution",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["step_resolution"],
    type=int,
    show_default=True,
    help="simulation step resolution in seconds",
)
@click.option(
    "-o",
    "--output",
    default=DEFAULT_TIME_SERIES_IMPACT_ANALYSIS_PARAMS["output_dir"],
    show_default=True,
    help="output directory",
)
@click.pass_context
def time_series(
    ctx, feeders, force, start, end, pv_profile, resolution, output
):
    """Transform input data for a time series simulation"""
    input_path = ctx.parent.params["input_path"]
    handle_existing_dir(output, force)
    simulation_params = {
        "start_time": start,
        "end_time": end,
        "step_resolution": resolution,
        "simulation_type": SimulationType.QSTS,
    }
    EpriModel.transform(
        input_path=input_path,
        output_path=output,
        simulation_params=simulation_params,
        simulation_model=TimeSeriesAnalysisModel,
        feeders=feeders,
        pv_profile=pv_profile,
    )
    print(
        f"Transformed data from {input_path} to {output} for TimeSeries Analysis."
    )


class EpriModel(BaseOpenDssModel):
    """EPRI Feeder Model Inputs Class"""

    TRANSFORM_SUBCOMMANDS = {
        "snapshot": snapshot,
        "time-series": time_series,
    }
    MASTER_FILENAME_BY_FEEDER = {
        "J1": "Master_noPV.dss",
        "K1": "Master_NoPV.dss",
        "M1": "Master_NoPV.dss",
    }

    def __init__(self, data):
        self._feeder = data.pop("feeder")
        self._loadshape_directory = data.pop("loadshape_directory")
        self._opendss_directory = data.pop("opendss_directory")
        self._metadata_directory = data.pop("metadata_directory", None)
        self._master = data.pop("master")
        self._name = data.pop("name")
        self._pv_locations = data.pop("pv_locations")

    @property
    def dc_ac_ratio(self):
        return 1.15

    @property
    def substation(self):
        return None

    @property
    def feeder(self):
        return self._feeder

    @property
    def loadshape_directory(self):
        return self._loadshape_directory

    @property
    def opendss_directory(self):
        return self._opendss_directory

    @property
    def master_file(self):
        return os.path.join(self._opendss_directory, self._master)

    @property
    def metadata_directory(self):
        return self._metadata_directory

    @property
    def name(self):
        return self._name

    @property
    def pv_locations(self):
        if not self._pv_locations:
            return []

        files = []
        for pv_location in self._pv_locations:
            _file = os.path.join(self._opendss_directory, pv_location)
            if not os.path.exists(_file):
                raise FileNotFoundError(f"File not exist - {_file}")
            files.append(_file)

        return files

    @property
    def pydss_controllers(self):
        return PyDSSControllerModel(
            controller_type=ControllerType.PV_CONTROLLER,
            name="volt_var_ieee_1547_2018_catB",
        )

    @staticmethod
    def get_transform_subcommand(name):
        if name not in EpriModel.TRANSFORM_SUBCOMMANDS:
            raise InvalidParameter(f"{name} is not supported")
        return EpriModel.TRANSFORM_SUBCOMMANDS[name]

    @staticmethod
    def list_transform_subcommands():
        return sorted(list(EpriModel.TRANSFORM_SUBCOMMANDS.keys()))

    @classmethod
    def transform(
        cls,
        input_path,
        output_path,
        simulation_params,
        simulation_model,
        feeders=("all",),
        existing_pv=True,
        pv_profile=None,
    ):
        config = []
        os.makedirs(output_path, exist_ok=True)

        if feeders == ("all",):
            feeders = [
                x
                for x in os.listdir(input_path)
                if os.path.isdir(os.path.join(input_path, x))
            ]

        for i, feeder in enumerate(feeders):
            pv_locations = []
            pv_path = os.path.join(input_path, feeder, "ExistingPV.dss")
            if existing_pv and os.path.exists(pv_path):
                pv_locations.append("ExistingPV.dss")
            master_filename = cls.MASTER_FILENAME_BY_FEEDER[feeder]
            name = f"{feeder}__deployment{i + 1}"
            data = {
                "feeder": feeder,
                "loadshape_directory": None,
                "opendss_directory": os.path.join(input_path, feeder),
                "master": master_filename,
                "name": name,
                "pv_locations": pv_locations,
            }
            model = cls(data)
            path = os.path.join(output_path, feeder)
            deployment = model.create_deployment(
                name,
                path,
                pv_profile=pv_profile,
                hierarchy=SimulationHierarchy.FEEDER,
            )

            item = {
                "deployment": deployment,
                "simulation": simulation_params,
                "name": model.name,
                "model_type": simulation_model.__name__,
            }
            config.append(simulation_model.validate(item).dict())

        filename = os.path.join(output_path, SOURCE_CONFIGURATION_FILENAME)
        with open(filename, "w") as f_out:
            json.dump(config, f_out, indent=2, cls=ExtendedJSONEncoder)
        logger.info("Wrote config to %s", filename)



"""Analysis of PyDSS simulations."""

import logging
import math
import os
import re

import pandas as pd

from jade.exceptions import InvalidParameter, InvalidConfiguration
from jade.jobs.job_analysis import JobAnalysis
from jade.utils.utils import load_data
from PyDSS.pydss_results import PyDssResults
from disco.sources.gem.make_element_bus_mapping import get_bus_to_element, \
    REGION_BUS_MAPPING_FILENAME


logger = logging.getLogger(__name__)


class PyDssAnalysis(JobAnalysis):
    """Performs analysis of PyDSS simulations."""

    def read_results(self, simulation):
        """Return all results dataframes for a simulation.

        Parameters
        ----------
        simulation : str | PyDssSimulationBase

        Returns
        -------
        dict
            class name : PyDssResult instance

        """
        if isinstance(simulation, str):
            simulation = self.get_simulation(simulation)

        return PyDssResults(simulation.pydss_project_path)


class PyDssScenarioAnalysis:
    """Performs analysis on one PyDSS scenario results."""

    _REGEX_PHASE_ANY_TERMINAL_1 = re.compile(rf"[ABCN]1")

    def __init__(self, job, results, scenario_name=None):
        self._job = job
        self._feeder = job.model.deployment.feeder
        if not results.scenarios:
            raise InvalidParameter("there are no scenarios in the results")

        if scenario_name is None:
            self._scenario = results.scenarios[0]
        else:
            self._scenario = results.get_scenario(scenario_name)

    def get_pu_bus_voltage_magnitudes(self):
        """Return per-unit voltage magnitudes for each bus.

        Returns
        -------
        dict
            Maps bus name to list of voltages.

        """
        voltages = {}
        for name in self._scenario.list_element_names("Buses", "puVmagAngle"):
            df = self._scenario.get_dataframe(
                "Buses",
                "puVmagAngle",
                name,
                phase_terminal=self._REGEX_PHASE_ANY_TERMINAL_1,
                mag_ang="mag",
            )
            voltages[name] = [row.mean() for _, row in df.iterrows()]

        return voltages

    def get_line_loading_percentages(self, fmt="dataframe"):
        """Return line loading values as percents for all lines.

        Parameters
        ----------
        fmt : str
            Controls output format. Must be 'dataframe' or 'json'."

        Returns
        -------
        dict
            Maps line name to loading percentages.

        """
        if fmt.lower() not in ("dataframe", "json"):
            raise InvalidParameter("fmt must be 'dataframe' or 'json'")

        loadings = {}
        columns = ["Line Loading (%)"]

        for name in self._scenario.list_element_names("Lines", "Currents"):
            df = self._scenario.get_dataframe(
                "Lines",
                "Currents",
                name,
                phase_terminal=self._REGEX_PHASE_ANY_TERMINAL_1
            )
            normal_amps = self._scenario.get_dataframe(
                "Lines",
                "NormalAmps",
                name,
            )
            values = []
            for i, row in df.iterrows():
                val = max(row.apply(lambda x: math.sqrt(x.real**2 + x.imag**2)))
                normal_amps_val = normal_amps.loc[i][normal_amps.columns[0]]
                values.append(val / normal_amps_val * 100)

            df = pd.DataFrame(values, index=df.index, columns=columns)
            if fmt == "json":
                loadings[name] = df.to_json(orient="records")
            else:
                loadings[name] = df

        return loadings

    def get_transformer_loading_percentages(self, fmt="dataframe"):
        """Return transformer loading values as percents for all transformers.

        Parameters
        ----------
        fmt : str
            Controls output format. Must be 'dataframe' or 'json'."

        Returns
        -------
        dict
            Maps transformer name to loading percentages pd.DataFrame.

        """
        if fmt.lower() not in ("dataframe", "json"):
            raise InvalidParameter("fmt must be 'dataframe' or 'json'")

        #transformers = self._scenario.read_element_info_file("Transformers")
        #transformers_phase_info = self._scenario.read_element_info_file("TransformersPhase")

        loadings = {}
        columns = ["Transformer Loading (%)"]
        for name in self._scenario.list_element_names("Transformers", "Currents"):
            #short_name = name.replace("Transformer.", "")
            df = self._scenario.get_dataframe(
                "Transformers",
                "Currents",
                name,
                phase_terminal=self._REGEX_PHASE_ANY_TERMINAL_1
            )
            normal_amps = self._scenario.get_dataframe(
                "Transformers",
                "NormalAmps",
                name,
            )

            # TODO: this needs some minor correction

            #transformer = transformers.query(f"Name == '{short_name}'").iloc[0]
            #transformer_phase = transformers_phase_info.query(f"Transformer == '{name}'").iloc[0]
            #num_windings = transformer["NumWindings"]
            #high_side_connection = transformer_phase["HighSideConnection"]
            #num_phases = len(df.columns)
            #if num_phases == 3 and high_side_connection == "wye":
            #    phase = 4
            #elif num_phases == 3 and high_side_connection == "delta":
            #    phase = 3
            #else:
            #    phase = num_phases

            values = []
            for i, row in df.iterrows():
                val = max(row.apply(lambda x: math.sqrt(x.real**2 + x.imag**2)))
                normal_amps_val = normal_amps.loc[i][normal_amps.columns[0]]
                values.append(val / normal_amps_val * 100)

            df = pd.DataFrame(values, index=df.index, columns=columns)
            if fmt == "json":
                loadings[name] = df.to_json(orient="records")
            else:
                loadings[name] = df

        return loadings

    def get_kw_at_bus_mapping(self):
        """Return a mapping of bus to PV system and load kW values.

        Requires that the script make_element_bus_mapping.py has been run on this feeder.

        Returns
        -------
        dict
            Key is bus name. Value is a list of dicts containing element info
            and kW values.

            Example: {"123456": [{"type": "pv_systems", "name": "pv_1234", "kW", 1.3}]}

        """
        input_directory = self._job.model.deployment.directory
        bus_mapping_file = os.path.join(
            input_directory, REGION_BUS_MAPPING_FILENAME
        )
        if not os.path.exists(bus_mapping_file):
            raise InvalidConfiguration(
                f"{bus_mapping_file} does not exist. Has make_element_bus_mapping.py been run?"
            )

        summary = load_data(bus_mapping_file)
        if self._feeder not in summary:
            raise InvalidConfiguration(
                f"{bus_mapping_file} does not contain feeder={self._feeder}"
            )

        feeder_file = summary[self._feeder]
        feeder_mapping = load_data(feeder_file)

        bus_to_elems = {}
        get_bus_to_element(bus_to_elems, feeder_mapping, "pv_systems")
        get_bus_to_element(bus_to_elems, feeder_mapping, "loads")

        pv_systems = self._scenario.read_element_info_file("PVSystems")
        loads = self._scenario.read_element_info_file("Loads")
        for elements in bus_to_elems.values():
            to_delete = []
            for i, element in enumerate(elements):
                src = pv_systems if element["type"] == "pv_systems" else loads
                df = src[src["Name"] == element["name"]]
                if len(df) == 0:
                    # The data may include other PV systems.
                    to_delete.append(i)
                    continue
                element["kW"] = df["kW"].values[0]
            for i in reversed(to_delete):
                elements.pop(i)

        return bus_to_elems



"""Defines deployment parameters for auto-generated scenarios."""

from collections import namedtuple
import logging

from jade.jobs.job_parameters_interface import JobParametersInterface
from jade.common import DEFAULT_SUBMISSION_GROUP
from disco.models.factory import make_model
from disco.models.base import ImpactAnalysisBaseModel
from disco.models.snapshot_impact_analysis_model import SnapshotImpactAnalysisModel
from disco.models.time_series_analysis_model import TimeSeriesAnalysisModel
from disco.models.upgrade_cost_analysis_model import UpgradeCostAnalysisModel

logger = logging.getLogger(__name__)


class DeploymentParameters(JobParametersInterface):
    """Represents deployment parameters for auto-generated scenarios."""

    parameters_type = namedtuple("DeploymentKeys", "name")
    DEFAULT_STEP_RESOLUTION = 900
    _EXTENSIONS = {
        SnapshotImpactAnalysisModel: "pydss_simulation",
        TimeSeriesAnalysisModel: "pydss_simulation",
        UpgradeCostAnalysisModel: "upgrade_simulation"
    }

    def __init__(self, estimated_run_minutes=None, **kwargs):
        self._estimated_run_minutes = estimated_run_minutes
        self._model = make_model(kwargs)
        self._submission_group = DEFAULT_SUBMISSION_GROUP

    def __repr__(self):
        return self.name

    @property
    def estimated_run_minutes(self):
        return self._estimated_run_minutes

    @estimated_run_minutes.setter
    def estimated_run_minutes(self, val):
        self._estimated_run_minutes = val

    @staticmethod
    def list_extensions():
        # this may need to be extended to all JobParametersInterface classes
        return list(DeploymentParameters._EXTENSIONS)

    @property
    def extension(self):
        return self._EXTENSIONS[type(self._model)]

    @property
    def model(self):
        """Return the input data model for the job."""
        return self._model

    @property
    def base_case(self):
        """Return the base case or None."""
        if isinstance(self._model, ImpactAnalysisBaseModel):
            return self._model.base_case
        return None

    def has_pydss_controllers(self):
        """Return True if pydss controllers are present."""
        if self._model.deployment.pydss_controllers is None:
            return False
        if isinstance(self._model.deployment.pydss_controllers, list):
            return bool(self._model.deployment.pydss_controllers)
        return True

    @property
    def name(self):
        return self._model.name

    @property
    def feeder(self):
        """Return the job's feeder."""
        return self._model.deployment.feeder

    def serialize(self):
        """Serialize deployment into a dictionary."""
        data = self._model.dict()
        data["extension"] = self.extension
        data["estimated_run_minutes"] = self.estimated_run_minutes
        return data

    @classmethod
    def deserialize(cls, data):
        return cls(**data)

    def add_blocking_job(self, name):
        self._model.blocked_by.add(name)

    def add_blocking_jobs(self, names):
        self._model.blocked_by.update(names)

    def get_blocking_jobs(self):
        return self._model.blocked_by

    def remove_blocking_job(self, name):
        self._model.blocked_by.remove(name)

    def set_blocking_jobs(self, blocking_jobs):
        self._model.blocked_by = blocking_jobs

    @property
    def cancel_on_blocking_job_failure(self):
        return True

    @property
    def submission_group(self):
        return self._submission_group

    @submission_group.setter
    def submission_group(self, group):
        self._submission_group = group



import logging
import os

from PyDSS.common import SnapshotTimePointSelectionMode
from PyDSS.reports.pv_reports import PF1_SCENARIO, CONTROL_MODE_SCENARIO

import disco
from disco.enums import SimulationType, AnalysisType
from disco.pipelines.enums import TemplateSection
from disco.pipelines.base import PipelineCreatorBase
from disco.pydss.common import TIME_SERIES_SCENARIOS
from disco.pydss.pydss_configuration_base import get_default_exports_file
from jade.models.pipeline import PipelineConfig
from jade.utils.utils import dump_data

logger = logging.getLogger(__name__)


class SnapshotPipelineCreator(PipelineCreatorBase):

    def create_pipeline(self, config_file):
        """Make snapshot pipeline config file"""
        stages = [self.make_simulation_stage()]
        if self.template.contains_postprocess():
            stages.append(self.make_postprocess_stage())

        config = PipelineConfig(stages=stages, stage_num=1)
        with open(config_file, "w") as f:
            f.write(config.json(indent=2))
        logger.info("Created pipeline config file - %s", config_file)

    def make_model_transform_command(self):
        options = self.template.get_transform_options(TemplateSection.MODEL)
        command = f"disco transform-model {self.template.inputs} snapshot {options}"
        logger.info("Make command - '%s'", command)
        return command

    def make_disco_config_command(self, section):
        # self.template reads the file fresh every time.
        # TODO: change the rest of this file to not constantly re-read it.
        template = self.template
        if template.preconfigured:
            model_inputs = template.inputs
        else:
            model_inputs = template.get_model_transform_output()
        options = template.get_config_options(section)
        # The code that builds the option string excludes False values but we need it here.
        if not template.get_config_params(TemplateSection.SIMULATION)["with_loadshape"]:
            options += " --no-with-loadshape"
        reports_filename = "generated_snapshot_reports.toml"
        dump_data(template.reports, reports_filename)
        exports_filename = get_default_exports_file(
            SimulationType.SNAPSHOT,
            AnalysisType(template.analysis_type),
        )
        command = (
            f"disco config snapshot {model_inputs} "
            f"--reports-filename={reports_filename} --exports-filename={exports_filename} {options}"
        )
        return command

    def make_prescreen_create_command(self):
        pass

    def make_prescreen_filter_command(self):
        pass

    def make_postprocess_command(self):
        commands = []
        impact_analysis = self.template.analysis_type == AnalysisType.IMPACT_ANALYSIS.value
        hosting_capacity = self.template.analysis_type == AnalysisType.HOSTING_CAPACITY.value
        if impact_analysis or hosting_capacity:
            # Postprocess to make summary tables
            inputs = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
            commands.append(f"disco make-summary-tables {inputs}")
            
            # Postprocess to compute hosting capacity
            if hosting_capacity:
                config_params = self.template.get_config_params(TemplateSection.SIMULATION)
                with_loadshape = config_params["with_loadshape"]
                auto_select_time_points = config_params["auto_select_time_points"]
                pf1 = config_params["pf1"]
                base_cmd = f"disco compute-hosting-capacity {inputs}"
                plot_cmd = f"disco plot {inputs}"
                scenarios = [CONTROL_MODE_SCENARIO]
                if pf1:
                    scenarios.append(PF1_SCENARIO)
                if with_loadshape and auto_select_time_points:
                    for scenario in scenarios:
                        for mode in SnapshotTimePointSelectionMode:
                            if mode != SnapshotTimePointSelectionMode.NONE:
                                commands.append(f"{base_cmd} --scenario={scenario} --time-point={mode.value}")
                else:
                    for scenario in scenarios:
                        commands.append(f"{base_cmd} --scenario={scenario}")
            
                # Plot
                for scenario in scenarios:
                    commands.append(f"{plot_cmd} --scenario {scenario}")

            # Postprocess to ingest results into sqlite database
            task_name = self.template.data["task_name"]
            if os.path.isabs(self.template.database):
                database = self.template.database
            else:
                database = os.path.join(inputs, self.template.database)
            model_inputs = self.template.inputs
            commands.append(
                'disco ingest-tables '
                f'--task-name "{task_name}" '
                f'--database {database} '
                f'--model-inputs {model_inputs} '
                f'{inputs}'
            )

        return "\n".join(commands)


class TimeSeriesPipelineCreator(PipelineCreatorBase):
    """Time-series pipeline creator class"""

    def create_pipeline(self, config_file):
        """Make time-series pipeline config file"""
        stages = []
        if self.template.contains_prescreen():
            stages.append(self.make_prescreen_stage())
        stages.append(self.make_simulation_stage())
        if self.template.contains_postprocess():
            stages.append(self.make_postprocess_stage())

        config = PipelineConfig(stages=stages, stage_num=1)
        with open(config_file, "w") as f:
            f.write(config.json(indent=2))
        logger.info("Created pipeline config file - %s", config_file)

    def make_model_transform_command(self):
        options = self.template.get_transform_options(TemplateSection.MODEL)
        command = f"disco transform-model {self.template.inputs} time-series {options}"
        logger.info("Make command - '%s'", command)
        return command

    def make_disco_config_command(self, section):
        if self.template.preconfigured:
            model_inputs = self.template.inputs
        else:
            model_inputs = self.template.get_model_transform_output()
        options = self.template.get_config_options(section)
        reports_filename = "generated_time_series_reports.toml"
        dump_data(self.template.reports, reports_filename)
        command = (
            f"disco config time-series {model_inputs} "
            f"--reports-filename={reports_filename} {options}"
        )
        if self.template.analysis_type == AnalysisType.COST_BENEFIT.value:
            # These must not be user-configurable and don't go in the template.
            command += " --feeder-losses=false --thermal-metrics=false --voltage-metrics=false"

        logger.info("Make command - '%s'", command)
        return command

    def make_prescreen_create_command(self):
        config_params = self.template.get_config_params(TemplateSection.PRESCREEN)
        config_file = config_params["config_file"]
        prescreen_params = self.template.get_prescreen_params(TemplateSection.PRESCREEN)
        command = (
            f"disco prescreen-pv-penetration-levels {config_file} "
            f"create --config-file={prescreen_params['prescreen_config_file']}"
        )
        logger.info("Make command - '%s'", command)
        return command

    def make_prescreen_filter_command(self):
        config_params = self.template.get_config_params(TemplateSection.PRESCREEN)
        config_file = config_params["config_file"]

        prescreen_params = self.template.get_prescreen_params(TemplateSection.PRESCREEN)
        prescreen_output = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
        command = (
            f"disco prescreen-pv-penetration-levels {config_file} "
            f"filter-config {prescreen_output} "
            f"--config-file={prescreen_params['filtered_config_file']}"
        )
        logger.info("Make command - '%s'", command)
        return command

    def make_postprocess_command(self):
        commands = []
        impact_analysis = self.template.analysis_type == AnalysisType.IMPACT_ANALYSIS.value
        hosting_capacity = self.template.analysis_type == AnalysisType.HOSTING_CAPACITY.value
        if impact_analysis or hosting_capacity:
            inputs = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
            commands.append(f"disco make-summary-tables {inputs}")
            if hosting_capacity:
                for scenario in TIME_SERIES_SCENARIOS:
                    commands.append(f"disco compute-hosting-capacity {inputs} --scenario={scenario}")
                
                for scenario in TIME_SERIES_SCENARIOS:
                    commands.append(f"disco plot {inputs} --scenario={scenario}")
            
            # Postprocess to ingest results into sqlite database
            task_name = self.template.data["task_name"]
            
            if os.path.isabs(self.template.database):
                database = self.template.database
            else:
                database = os.path.join(inputs, self.template.database)
            model_inputs = self.template.inputs
            commands.append(
                'disco ingest-tables '
                f'--task-name "{task_name}" '
                f'--database {database} '
                f'--model-inputs {model_inputs} '
                f'{inputs}'
            )

        elif self.template.analysis_type == AnalysisType.COST_BENEFIT.value:
            inputs = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
            commands.append(f"disco-internal make-cba-tables {inputs}")
        
        return "\n".join(commands)


class UpgradePipelineCreator(PipelineCreatorBase):
    """Upgrade pipeline creator class"""

    def create_pipeline(self, config_file):
        stages = [self.make_simulation_stage()]
        if self.template.contains_postprocess():
            stages.append(self.make_postprocess_stage())
        
        config = PipelineConfig(stages=stages, stage_num=1)
        with open(config_file, "w") as f:
            f.write(config.json(indent=2))
        logger.info("Created pipeline config file - %s", config_file)
    
    def make_model_transform_command(self):
        options = self.template.get_transform_options(TemplateSection.MODEL)
        command = f"disco transform-model {self.template.inputs} upgrade {options}"
        logger.info("Make command - '%s'", command)
        return command
    
    def make_prescreen_create_command(self):
        pass

    def make_prescreen_filter_command(self):
        pass

    def make_disco_config_command(self, section):
        if self.template.preconfigured:
            model_inputs = self.template.inputs
        else:
            model_inputs = self.template.get_model_transform_output()
        
        options = self.template.get_config_options(section)
        command = f"disco config upgrade {model_inputs} {options}"
        logger.info("Make command - '%s'", command)
        return command

    def make_postprocess_command(self):
        inputs = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
        command = f"disco-internal make-upgrade-tables {inputs}"
        return command


class UpgradePipelineCreator(PipelineCreatorBase):
    """Upgrade pipeline creator class"""

    def create_pipeline(self, config_file):
        stages = [self.make_simulation_stage()]
        if self.template.contains_postprocess():
            stages.append(self.make_postprocess_stage())
        
        config = PipelineConfig(stages=stages, stage_num=1)
        with open(config_file, "w") as f:
            f.write(config.json(indent=2))
        logger.info("Created pipeline config file - %s", config_file)
    
    def make_model_transform_command(self):
        options = self.template.get_transform_options(TemplateSection.MODEL)
        command = f"disco transform-model {self.template.inputs} upgrade {options}"
        logger.info("Make command - '%s'", command)
        return command
    
    def make_prescreen_create_command(self):
        pass

    def make_prescreen_filter_command(self):
        pass

    def make_disco_config_command(self, section):
        if self.template.preconfigured:
            model_inputs = self.template.inputs
        else:
            model_inputs = self.template.get_model_transform_output()
        
        options = self.template.get_config_options(section)
        command = f"disco config upgrade {model_inputs} {options}"
        logger.info("Make command - '%s'", command)
        return command

    def make_postprocess_command(self):
        inputs = os.path.join("$JADE_PIPELINE_OUTPUT_DIR", f"output-stage{self.stage_num-1}")
        command = f"disco-internal make-upgrade-tables {inputs}"
        return command



"""
Goal is to take the PyDSS hd5 store the compute hosting capacity analysis
"""

# python standard imports
from pathlib import Path
import os
import json
import copy
import logging
from textwrap import indent

# python third-party imports
import numpy as np

# python internal imports
from PyDSS.pydss_results import PyDssResults
from jade.common import CONFIG_FILE, JOBS_OUTPUT_DIR
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.jobs.results_aggregator import ResultsAggregator
from jade.utils.utils import dump_data, load_data
from jade.loggers import setup_logging


# setup logger
logger = logging.getLogger(__name__)


def compute_hosting_capacity_for_pydss(
    output_dir: str,
    output_json: str = "./hosting-capacity-from-pydss-results.json",
    threshold_voltage: float = 1.05,
    threshold_overloading: float = 100.0,
):

    setup_logging(__name__, None, console_level=logging.INFO, packages=["disco"])

    output_path = Path(output_dir)
    orig = Path(os.getcwd())
    try:
        os.chdir(output_path.parent)
        _compute_hosting_capacity_for_pydss(
            output_path, output_json, threshold_voltage, threshold_overloading
        )
    finally:
        os.chdir(orig)


def _compute_hosting_capacity_for_pydss(
    output_path, output_json: str, threshold_voltage: float, threshold_overloading
):

    """Parse all the jobs in JADE output directory and compute hosting capacity for each
    feeders by scenario and by time step taking pydss results"""

    config = create_config_from_file(output_path / CONFIG_FILE)
    jobs = []
    results = ResultsAggregator.list_results(output_path)
    result_lookup = {x.name: x for x in results}
    for job in config.iter_pydss_simulation_jobs():
        if job.name not in result_lookup:
            logger.info("Skip missing job %s", job.name)
            continue
        if result_lookup[job.name].return_code != 0:
            logger.info("Skip failed job %s", job.name)
            continue
        if job.model.is_base_case:
            continue
        jobs.append(job)

    # Container for storing max voltage and overloadings
    violation_container = []

    # Loop through all the jobs
    for job in jobs:

        job_path = output_path / JOBS_OUTPUT_DIR / job.name / "pydss_project"
        results = PyDssResults(job_path)

        substation = job.model.deployment.substation
        sample = job.model.deployment.project_data.get("sample", "-1")
        placement = job.model.deployment.project_data.get("placement", "-1")
        pen_level = job.model.deployment.project_data.get("penetration_level", "-1")

        for scenario in results.scenarios:

            df_buses = scenario.get_full_dataframe("Circuits", "AllBusMagPu")
            df_loadings = scenario.get_full_dataframe("CktElement", "ExportLoadingsMetric")

            violations = []
            max_thermal_dict = df_loadings.max(axis=1).to_dict()
            for key, value in df_buses.max(axis=1).to_dict().items():

                violations.append(
                    {
                        "timestamp": key.strftime("%Y-%m-%d %H:%M:%S"),
                        "max_voltage": value,
                        "penetration_level": float(pen_level),
                        "max_loading": max_thermal_dict[key],
                    }
                )

            violation_container.append(
                {
                    "substation": substation,
                    "scenario": scenario._name,
                    "placement_sample": str(placement) + "__" + str(sample),
                    "violations": violations,
                }
            )

    hosting_capacity_result = []
    substations = set(map(lambda x: x["substation"], violation_container))

    for substation in substations:

        substation_violations = [
            el for el in violation_container if el["substation"] == substation
        ]

        scenarios = set(map(lambda x: x["scenario"], substation_violations))
        for scenario in scenarios:

            scenario_violations = [
                el["violations"] for el in substation_violations if el["scenario"] == scenario
            ]

            timestamps = set(map(lambda x: x["timestamp"], scenario_violations[0]))

            for timestamp in timestamps:

                timestamp_violations = [
                    el
                    for sub_array in scenario_violations
                    for el in sub_array
                    if el["timestamp"] == timestamp
                ]

                max_voltages = np.array([el["max_voltage"] for el in timestamp_violations])
                max_loadings = np.array([el["max_loading"] for el in timestamp_violations])
                pen_levels = np.array([el["penetration_level"] for el in timestamp_violations])

                v_hc = (
                    np.max(pen_levels[max_voltages < threshold_voltage])
                    if len(pen_levels[max_voltages < threshold_voltage])
                    else None
                )

                t_hc = (
                    np.max(pen_levels[max_loadings < threshold_overloading])
                    if len(pen_levels[max_loadings < threshold_overloading])
                    else None
                )

                vt_hc = (
                    min(filter(lambda x: x is not None, [v_hc, t_hc]))
                    if any([v_hc, t_hc])
                    else None
                )

                hosting_capacity_result.append(
                    {
                        "substation": substation,
                        "scenario": scenario,
                        "timestamp": timestamp,
                        "hosting_capacity": {
                            "overvoltage_hc": v_hc,
                            "thermal_hc": t_hc,
                            "overvoltage_thermal_hc": vt_hc,
                        },
                    }
                )

    dump_data({"hc_result": hosting_capacity_result}, output_json, indent=2)
    logger.info(f"Written file named {output_json}")


if __name__ == "__main__":
    pass



"""Download EPRI source data of J1, K1, and M1 feeders."""

import logging
import os
import shutil
import tempfile
import zipfile
from urllib.parse import urlparse
from urllib.request import urlopen

from jade.utils.utils import dump_data


logger = logging.getLogger(__name__)


EPRI_FEEDER_MODELS = {
    "J1": "https://dpv.epri.com/media/feeder_j1_opendss_model.zip",
    "K1": "https://dpv.epri.com/media/feeder_k1_opendss_model.zip",
    "M1": "https://dpv.epri.com/media/feeder_m1_opendss_model.zip",
}


def download_epri_feeder_opendss_data(feeder, output_directory=None):
    """Download EPRI feeder OpenDSS model from website.

    Parameters
    ----------
    feeder : str
        The feeder name
    outdir : str
        The output directory for downloaded OpenDSS model.

    Returns
    -------
    str
        the directory of downloaded OpenDSS model.

    Raises
    ------
    ValueError
        Invalid feeder, choose from - 'J1', 'K1', 'M1'.
    """
    if feeder not in EPRI_FEEDER_MODELS:
        raise ValueError("Invalid feeder, choose from - 'J1', 'K1', 'M1'.")

    output_directory = output_directory or tempfile.gettempdir()
    opendss_directory = os.path.join(output_directory, f"{feeder}")

    if os.path.exists(opendss_directory):
        return opendss_directory

    os.makedirs(opendss_directory)
    try:
        url = EPRI_FEEDER_MODELS[feeder]
        # download zip file
        file_name = os.path.basename(urlparse(url).path)
        zip_file = os.path.join(opendss_directory, file_name)
        with urlopen(url) as response, open(zip_file, "wb") as f:
            shutil.copyfileobj(response, f)
        # extract zip file
        with zipfile.ZipFile(zip_file) as f:
            for info in f.namelist():
                file_name = os.path.basename(info)
                if not file_name:
                    continue
                dst_file = os.path.join(opendss_directory, file_name)
                with f.open(info) as src, open(dst_file, "wb") as dst:
                    shutil.copyfileobj(src, dst)
        os.remove(zip_file)
    except Exception:
        shutil.rmtree(opendss_directory)
        raise

    return opendss_directory


def define_epri_source_format(output_directory=None):
    data = {
        "type": "EpriModel"
    }
    
    if output_directory and os.path.exists(output_directory):
        filename = os.path.join(output_directory, "format.toml")
    else:
        filename = "format.toml"
    
    dump_data(data, filename)



import logging
import os

import click

from jade.loggers import setup_logging
from disco.pipelines.utils import ensure_jade_pipeline_output_dir
from disco.storage.db import create_database
from disco.storage.core import StoragePipeline


@click.command()
@click.argument("output")
@click.option(
    "-t", "--task-name",
    type=click.STRING,
    required=True,
    help="Name of the DISCO simulation/analysis task"
)
@click.option(
    "-m", "--model-inputs",
    type=click.Path(),
    required=True,
    help="Model inputs of the DISCO simulation/analysis task"
)
@click.option(
    "-n", "--notes",
    type=click.STRING,
    default=None,
    help="Notes about this DISCO simulation/analysis task"
)
@click.option(
    "-d", "--database",
    type=click.Path(),
    default="disco.sqlite",
    help="The path of new or existing SQLite database"
)
@click.option(
    "-v", "--verbose",
    is_flag=True,
    default=False,
    help="Enable debug logging"
)
def ingest_tables(
    output,
    task_name,
    model_inputs, 
    notes,
    database,
    verbose
):
    """Ingest DISCO simulation/analysis reports into SQLite database"""
    level = logging.DEBUG if verbose else logging.INFO
    logger = setup_logging(__name__, None, console_level=level, packages=["disco"])

    output = ensure_jade_pipeline_output_dir(output)
    database = ensure_jade_pipeline_output_dir(database)
    
    db_created = False
    if not os.path.exists(database):
        create_database(database)
        db_created = True
    
    try:
        data = {
            "output": os.path.abspath(output),
            "task_name": task_name,
            "model_inputs": os.path.abspath(model_inputs),
            "notes": notes or "",
            "database": database
        }
        pipeline = StoragePipeline(database=database)
        pipeline.run(data=data)
    except Exception as e:
        if db_created:
            os.remove(database)
        else:
            # TODO: rollback if any error
            pass
        raise e



"""Defines time series impact analysis model."""

from typing import Optional

from pydantic.v1 import Field

from .base import ImpactAnalysisBaseModel


class TimeSeriesAnalysisModel(ImpactAnalysisBaseModel):
    """Data model for time series impact analysis"""
    # Currently unused and just taking up space in config files.
    #include_pv_clipping: Optional[bool] = Field(
    #    title="include_pv_clipping",
    #    default=True,
    #    description="Whether to include PV clipping metrics",
    #)
    #include_pv_curtailment: Optional[bool] = Field(
    #    title="include_pv_curtailment",
    #    default=True,
    #    description="Whether to include PV curtailment metrics",
    #)
    #include_voltage_metrics: Optional[bool] = Field(
    #    title="include_voltage_metrics",
    #    default=True,
    #    description="Whether to include voltage metrics",
    #)
    #include_thermal_metrics: Optional[bool] = Field(
    #    title="include_thermal_metrics",
    #    default=True,
    #    description="Whether to include thermal metrics",
    #)
    #include_feeder_losses: Optional[bool] = Field(
    #    title="include_feeder_losses",
    #    default=True,
    #    description="Whether to include feeder loss metrics",
    #)
    #include_capacitor_state_change_count: Optional[bool] = Field(
    #    title="include_capacitor_state_change_count",
    #    default=True,
    #    description="Whether to include capacitor state change count metrics",
    #)
    #include_reg_control_state_change_count: Optional[bool] = Field(
    #    title="include_reg_control_state_change_count",
    #    default=True,
    #    description="Whether to include regulator state change count metrics",
    #)
    #granularity: Optional[str] = Field(
    #    title="granularity",
    #    default="per_element_per_time_point",
    #    description="Granularity of data to collect",
    #)

    class Config:
        title = "TimeSeriesAnalysisModel"
        anystr_strip_whitespace = True
        validate_assignment = True



"""Runs a simulation through PyDSS."""

import fileinput
import logging
import os
import re
import shutil
from pathlib import Path

from jade.common import OUTPUT_DIR
from jade.utils.utils import modify_file, interpret_datetime
from PyDSS.common import DATE_FORMAT

from disco.enums import SimulationType
from disco.pydss.common import ConfigType
from disco.pydss.pydss_simulation_base import PyDssSimulationBase


logger = logging.getLogger(__name__)

DEFAULTS = {
    "output": OUTPUT_DIR,
}


class PyDssSimulation(PyDssSimulationBase):
    """Runs a PyDss simulation."""

    SIMULATION_TYPE_TO_PYDSS_TYPE = {
        SimulationType.SNAPSHOT: "Snapshot",
        SimulationType.QSTS: "QSTS",
        SimulationType.TIME_SERIES: "QSTS",
    }

    def __init__(self, pydss_inputs, job, output=DEFAULTS["output"]):
        """Constructs a PyDssSimulation.

        Raises
        ------
        InvalidParameter
            Raised if any parameter is invalid.

        """
        super(PyDssSimulation, self).__init__(
            pydss_inputs,
            job,
            output
        )

    def __repr__(self):
        return f"Simulation: {self._model}"

    @classmethod
    def create(cls, pydss_inputs, job, output):
        logger.debug("params=%s", job)
        sim = cls(pydss_inputs, job, output=output)
        return sim

    @staticmethod
    def generate_command(job, output, config_file, verbose=False):
        command = [
            "jade-internal run pydss_simulation",
            f"--name={job.name}",
            f"--output={output}",
            f"--config-file={config_file}",
        ]

        if verbose:
            command.append("--verbose")

        return " ".join(command)

    def _get_scenario_names(self):
        return [x["name"] for x in self._pydss_inputs[ConfigType.SCENARIOS]]

    def _get_deployment_input_path(self):
        return os.path.join(
            self._dss_dir,
            self._get_deployment_input_filename(),
        )

    @staticmethod
    def _get_deployment_input_filename():
        return "deployment.dss"

    @staticmethod
    def _is_dss_file_path_absolute():
        return False

    def _modify_open_dss_parameters(self):
        deployment_filename = self._get_deployment_input_path()
        shutil.copyfile(self._model.deployment.deployment_file,
                        deployment_filename)
        modify_file(deployment_filename, self._recalculate_kva)
        logger.info("Modified kVA in %s", deployment_filename)

        modify_file(deployment_filename, self.fix_redirects, Path(deployment_filename))
        logger.info("Modified redirect path in %s", deployment_filename)

    @staticmethod
    def _minutes_from_midnight(timetuple):
        return timetuple.tm_hour * 60 + timetuple.tm_min

    def _modify_pydss_simulation_params(self, config):
        start_time = self._model.simulation.start_time
        end_time = self._model.simulation.end_time
        if start_time is not None:
            config["start_time"] = start_time.strftime(DATE_FORMAT)

        match = self._REGEX_YEAR.search(config["start_time"])
        assert match, config["start_time"]
        start_time_year = int(match.group(1))
        self._make_years_match(start_time_year, config, "loadshape_start_time")

        if end_time is not None:
            config["simulation_duration_min"] = (end_time - start_time).total_seconds() / 60

        if self._model.simulation.step_resolution is not None:
            config["step_resolution_sec"] = self._model.simulation.step_resolution

        config["simulation_type"] = self.SIMULATION_TYPE_TO_PYDSS_TYPE[self._model.simulation.simulation_type]

    def _recalculate_kva(self, line, *args, **kwargs):
        """Adjust kVA for dc_ac_ratio and kva_to_kw_rating."""
        values = []
        pmpp = None
        old_pct_pmpp = None
        pct_pmpp_index = None
        old_kva = None
        kva_index = None

        # Extract the kVA token, recalculate it, then add it back.
        for i, token in enumerate(line.split()):
            val_pair = token.split("=")
            if len(val_pair) == 2:
                if val_pair[0].lower() == "pmpp":
                    pmpp = float(val_pair[1])
                elif val_pair[0].lower() == "%pmpp":
                    old_pct_pmpp = float(val_pair[1])
                    pct_pmpp_index = i
                elif val_pair[0].lower() == "kva":
                    old_kva = float(val_pair[1])
                    kva_index = i
            values.append(token)

        if pmpp is None or old_kva is None:
            # No change required.
            return line

        dc_ac_ratio = self._model.deployment.dc_ac_ratio
        kva_to_kw_rating = self._model.deployment.kva_to_kw_rating

        if dc_ac_ratio is None or kva_to_kw_rating is None:
            assert dc_ac_ratio is None
            assert kva_to_kw_rating is None
            # No change required.
            return line

        kva = kva_to_kw_rating * pmpp / dc_ac_ratio
        values[kva_index] = f"kVA={kva}"
        logger.debug("old kVA = %s, new kVA = %s", old_kva, kva)

        if old_pct_pmpp is not None or self._add_pct_pmpp:
            pct_pmpp = self._irradiance_scaling_factor / dc_ac_ratio
            token = f"%Pmpp={pct_pmpp}"
            if pct_pmpp_index is None:
                values.append(token)
                logger.debug("Added pct_pmpp = %s", pct_pmpp)
            else:
                values[pct_pmpp_index] = token
                logger.debug("old pct_pmpp = %s, new pct_pmpp = %s",
                             old_pct_pmpp, pct_pmpp)

        return " ".join(values) + "\n"

    def fix_redirects(self, line, deployment_filename, **kwargs):
        """Fix the paths to a redirect file."""
        regex = re.compile(r"^([Rr]edirect\s+)(.*)")
        match = regex.search(line)
        if match is None:
            return line

        # The deployment file is now farther away from the base files.
        # This will change Redirect ../OpenDSS/Master.dss
        # to something like
        # Redirect ../../../../../model-inputs/X/OpenDSS/Master.dss

        base_path = Path(".")
        assert len(deployment_filename.parts) >= 2
        parts = [".."] * (len(deployment_filename.parts) - 1)
        parts.append(self._model.deployment.directory)
        base_path = Path(parts[0])
        for part in parts[1:]:
            base_path /= part

        path = Path(match.group(2))
        parts = [x for x in path.parts if str(x) != ".."]
        new_path = base_path
        for parent in parts:
            new_path /= parent
        new_line = match.group(1) + str(new_path)
        return new_line



"""Contains OpenDSS utility functions."""

import fileinput
import logging
import re

logger = logging.getLogger(__name__)


def read_capacitor_changes(event_log):
    """Read the capacitor state changes from an OpenDSS event log.

    Parameters
    ----------
    event_log : str
        Path to event log

    Returns
    -------
    dict
        Maps capacitor names to count of state changes.

    """
    capacitor_changes = {}
    regex = re.compile(r"(Capacitor\.\w+)")

    data = read_event_log(event_log)
    for row in data:
        match = regex.search(row["Element"])
        if match:
            name = match.group(1)
            if name not in capacitor_changes:
                capacitor_changes[name] = 0
            action = row["Action"].replace("*", "")
            if action in ("OPENED", "CLOSED", "STEP UP"):
                capacitor_changes[name] += 1

    return capacitor_changes


def read_event_log(filename):
    """Return OpenDSS event log information.

    Parameters
    ----------
    filename : str
        path to event log file.


    Returns
    -------
    list
        list of dictionaries (one dict for each row in the file)

    """
    data = []

    with open(filename) as f_in:
        for line in f_in:
            tokens = [x.strip() for x in line.split(",")]
            row = {}
            for token in tokens:
                name_and_value = [x.strip() for x in token.split("=")]
                name = name_and_value[0]
                value = name_and_value[1]
                row[name] = value
            data.append(row)

    return data


def comment_out_leading_strings(filename, strings):
    """Insert a comment character on any line that begins with one of strings.

    Parameters
    ----------
    filename : str
    strings : list

    """
    with fileinput.input(files=[filename], inplace=True) as f_in:
        for line in f_in:
            if line.lower().startswith(strings):
                line = "!" + line
            print(line, end="")
